import collections
from copy import copy
from pprint import pformat as prettyformat
from functools import partial
from itertools import chain
from pathlib import Path
import warnings
import gc

import xarray as xr
import animatplot as amp
from matplotlib import pyplot as plt
from matplotlib.animation import PillowWriter

from mpl_toolkits.axes_grid1 import make_axes_locatable

import numpy as np
from dask.diagnostics import ProgressBar

from .geometries import apply_geometry
from .plotting.animate import (
    animate_poloidal,
    animate_pcolormesh,
    animate_line,
    _add_controls,
    _normalise_time_coord,
    _parse_coord_option,
)
from .region import _from_region
from .utils import (
    _add_cartesian_coordinates,
    _get_bounding_surfaces,
    _split_into_restarts,
)


@xr.register_dataset_accessor("bout")
class BoutDatasetAccessor:
    """
    Contains BOUT-specific methods to use on BOUT++ datasets opened using
    `open_boutdataset()`.

    These BOUT-specific methods and attributes are accessed via the bout
    accessor, e.g. `ds.bout.options` returns a `BoutOptionsFile` instance.
    """

    def __init__(self, ds):
        self.data = ds
        self.metadata = ds.attrs.get("metadata")  # None if just grid file
        self.options = ds.attrs.get("options")  # None if no inp file

    def __str__(self):
        """
        String representation of the BoutDataset.

        Accessed by print(ds.bout)
        """

        styled = partial(prettyformat, indent=4, compact=True)
        text = (
            "<xbout.BoutDataset>\n"
            + "Contains:\n{}\n".format(str(self.data))
            + "Metadata:\n{}\n".format(styled(self.metadata))
        )
        if self.options:
            text += "Options:\n{}".format(self.options)
        return text

    # def __repr__(self):
    #    return 'boutdata.BoutDataset(', {}, ',', {}, ')'.format(self.datapath,
    #  self.prefix)

    def get_field_aligned(self, name, caching=True):
        """
        Get a field-aligned version of a variable, calculating (and caching in the
        Dataset) if necessary

        Parameters
        ----------
        name : str
            Name of the variable to get field-aligned version of
        caching : bool, optional
            Save the field-aligned variable in the Dataset (default: True)
        """
        aligned_name = name + "_aligned"
        try:
            result = self.data[aligned_name]
            if result.direction_y != "Aligned":
                raise ValueError(
                    aligned_name + " exists, but is not field-aligned, it "
                    "has direction_y=" + result.direction_y
                )
            return result
        except KeyError:
            if caching:
                self.data[aligned_name] = self.data[name].bout.to_field_aligned()
            return self.data[aligned_name]

    def to_field_aligned(self):
        """
        Create a new Dataset with all 3d variables transformed to field-aligned
        coordinates, which are shifted with respect to the base coordinates by an angle
        zShift
        """
        result = self.data.copy()

        for v in chain(result, result.coords):
            da = result[v]
            # Need to transform any z-dependent variables or coordinates
            if (result.metadata["bout_zdim"] in da.dims) and (
                da.attrs.get("direction_y", None) == "Standard"
            ):
                result[v] = da.bout.to_field_aligned()

        return result

    def from_field_aligned(self):
        """
        Create a new Dataset with all 3d variables transformed to non-field-aligned
        coordinates
        """
        result = self.data.copy()

        for v in chain(result, result.coords):
            da = result[v]
            # Need to transform any 3d variables or coordinates
            if (result.metadata["bout_zdim"] in da.dims) and (
                da.attrs.get("direction_y", None) == "Aligned"
            ):
                result[v] = da.bout.from_field_aligned()

        return result

    def from_region(self, name, with_guards=None):
        """
        Get a logically-rectangular section of data from a certain region.
        Includes guard cells from neighbouring regions.

        Parameters
        ----------
        name : str
            Region to get data for
        with_guards : int or dict of int, optional
            Number of guard cells to include, by default use MXG and MYG from BOUT++.
            Pass a dict to set different numbers for different coordinates.
        """
        return _from_region(self.data, name, with_guards)

    @property
    def _regions(self):
        if "regions" not in self.data.attrs:
            raise ValueError(
                "Called a method requiring regions, but these have not been created. "
                "Please set the 'geometry' option when calling open_boutdataset() to "
                "create regions."
            )
        return self.data.attrs["regions"]

    @property
    def fine_interpolation_factor(self):
        """
        The default factor to increase resolution when doing parallel interpolation
        """
        return self.data.metadata["fine_interpolation_factor"]

    @fine_interpolation_factor.setter
    def fine_interpolation_factor(self, n):
        """
        Set the default factor to increase resolution when doing parallel interpolation.

        Parameters
        -----------
        n : int
            Factor to increase parallel resolution by
        """
        ds = self.data
        ds.metadata["fine_interpolation_factor"] = n
        for da in ds.data_vars.values():
            da.metadata["fine_interpolation_factor"] = n

    def interpolate_parallel(self, variables, **kwargs):
        """
        Interpolate in the parallel direction to get a higher resolution version of a
        subset of variables.

        Note that the high-resolution variables are all loaded into memory, so most
        likely it is necessary to select only a small number. The toroidal_points
        argument can also be used to reduce the memory demand.

        Parameters
        ----------
        variables : str or sequence of str or ...
            The names of the variables to interpolate. If 'variables=...' is passed
            explicitly, then interpolate all variables in the Dataset.
        n : int, optional
            The factor to increase the resolution by. Defaults to the value set by
            BoutDataset.setupParallelInterp(), or 10 if that has not been called.
        toroidal_points : int or sequence of int, optional
            If int, number of toroidal points to output, applies a stride to toroidal
            direction to save memory usage. If sequence of int, the indexes of toroidal
            points for the output.
        method : str, optional
            The interpolation method to use. Options from xarray.DataArray.interp(),
            currently: linear, nearest, zero, slinear, quadratic, cubic. Default is
            'cubic'.

        Returns
        -------
        A new Dataset containing a high-resolution versions of the variables. The new
        Dataset is a valid BoutDataset, although containing only the specified variables.
        """

        if variables is ...:
            variables = [v for v in self.data]

        if isinstance(variables, str):
            variables = [variables]
        if isinstance(variables, tuple):
            variables = list(variables)

        # Need to start with a Dataset with attrs as merge() drops the attrs of the
        # passed-in argument.
        # Make sure the first variable has all dimensions so we don't lose any
        # coordinates
        def find_with_dims(first_var, dims):
            if first_var is None:
                dims = set(dims)
                for v in variables:
                    if set(self.data[v].dims) == dims:
                        first_var = v
                        break
            return first_var

        tcoord = self.data.metadata.get("bout_tdim", "t")
        zcoord = self.data.metadata.get("bout_zdim", "z")
        first_var = find_with_dims(None, self.data.dims)
        first_var = find_with_dims(first_var, set(self.data.dims) - set(tcoord))
        first_var = find_with_dims(first_var, set(self.data.dims) - set(zcoord))
        first_var = find_with_dims(
            first_var, set(self.data.dims) - set([tcoord, zcoord])
        )
        if first_var is None:
            raise ValueError(
                f"Could not find variable to interpolate with both "
                f"{self.data.metadata.get('bout_xdim', 'x')} and "
                f"{self.data.metadata.get('bout_ydim', 'y')} dimensions"
            )
        variables.remove(first_var)
        ds = self.data[first_var].bout.interpolate_parallel(
            return_dataset=True, **kwargs
        )
        xcoord = ds.metadata.get("bout_xdim", "x")
        ycoord = ds.metadata.get("bout_ydim", "y")
        for var in variables:
            da = self.data[var]
            if xcoord in da.dims and ycoord in da.dims:
                ds = ds.merge(
                    da.bout.interpolate_parallel(return_dataset=True, **kwargs)
                )
            elif ycoord not in da.dims:
                ds[var] = da
            # Can't interpolate a variable that depends on y but not x, so just skip

        # Apply geometry
        ds = apply_geometry(ds, ds.geometry)

        return ds

    def integrate_midpoints(self, variable, *, dims=None, cumulative_t=False):
        """
        Integrate using the midpoint rule for spatial dimensions, and trapezium rule for
        time.

        The quantity being integrated is assumed to be a scalar variable.

        When doing a 1d integral in the 'y' dimension, the integral is calculated as a
        poloidal integral if the variable is on the standard grid ('direction_y'
        attribute is "Standard"), or as a parallel-to-B integral if the variable is on
        the field-aligned grid ('direction_y' attribute is "Aligned").

        When doing a 2d integral over 'x' and 'y' dimensions, the integral will be over
        poloidal cross-sections if the variable is not field-aligned (direction_y ==
        "Standard") and over field-aligned surfaces if the variable is field-aligned
        (direction_ == "Aligned"). The latter seems unlikely to be useful as the
        surfaces depend on the arbitrary origin used for zShift.

        Is a method of BoutDataset accessor rather than of BoutDataArray so we can use
        other variables like `J`, `g11`, `g_22` for the integration.

        Note the xarray.DataArray.integrate() method uses the trapezium rule, which is
        not consistent with the way BOUT++ defines grid spacings as cell widths. Also,
        this way for example::

            inner = da.isel(x=slice(i)).bout.integrate_midpoints()
            outer = da.isel(x=slice(i, None).bout.integrate_midpoints()
            total = da.bout.integrate_midpoints()
            inner + outer == total

        while with the trapezium rule you would have to select ``radial=slice(i+1)`` for
        inner to get a similar relation to be true.

        Parameters
        ----------
        variable : str or DataArray
            Name of the variable to integrate, or the variable itself as a DataArray.
        dims : str, list of str or ...
            Dimensions to integrate over. Can be any combination of of the dimensions of
            the Dataset. Defaults to integration over all spatial dimensions. If `...`
            is passed, integrate over all dimensions including time.
        cumulative_t : bool, default False
            If integrating in time, return the cumulative integral (integral from the
            beginning up to each point in the time dimension) instead of the definite
            integral.
        """
        ds = self.data

        if isinstance(variable, str):
            variable = ds[variable]

        location = variable.cell_location
        suffix = "" if location == "CELL_CENTRE" else f"_{location}"

        tcoord = ds.metadata["bout_tdim"]
        xcoord = ds.metadata["bout_xdim"]
        ycoord = ds.metadata["bout_ydim"]
        zcoord = ds.metadata["bout_zdim"]

        if dims is None:
            dims = []
            if xcoord in ds.dims:
                dims.append(xcoord)
            if ycoord in ds.dims:
                dims.append(ycoord)
            if zcoord in ds.dims:
                dims.append(zcoord)
        elif dims is ...:
            dims = []
            if tcoord in ds.dims:
                dims.append(tcoord)
            if xcoord in ds.dims:
                dims.append(xcoord)
            if ycoord in ds.dims:
                dims.append(ycoord)
            if zcoord in ds.dims:
                dims.append(zcoord)
        elif isinstance(dims, str):
            dims = [dims]

        dx = ds[f"dx{suffix}"]
        dy = ds[f"dy{suffix}"]
        if ds.metadata["BOUT_VERSION"] >= 5.0:
            dz = ds[f"dz{suffix}"]
        else:
            dz = ds["dz"]

        # Work out the spatial volume element
        if xcoord in dims and ycoord in dims and zcoord in dims:
            # Volume integral, use the 3d Jacobian "J"
            spatial_volume_element = ds[f"J{suffix}"] * dx * dy * dz
        elif xcoord in dims and ycoord in dims:
            # 2d integral on poloidal planes
            if variable.direction_y == "Standard":
                # Need to use a metric constructed from basis vectors within the
                # poloidal plane, so use 'reciprocal basis vectors' Grad(x^i)
                # J = 1/sqrt(det(g_2d))
                # det(g_2d) = g11*g22 - g12**2
                g = ds[f"g11{suffix}"] * ds[f"g22{suffix}"] - ds[f"g12{suffix}"] ** 2
                J = 1.0 / np.sqrt(g)
            elif variable.direction_y == "Aligned":
                # Need to work out area element from metric coefficients. See book by
                # D'haeseleer, Hitchon, Callen and Shohet eq. (2.5.51).
                # Need to use a metric constructed from basis vectors within the
                # field-aligned x-y plane, so use 'tangent basis vectors' e_i
                # J = sqrt(g_11*g_22 - g_12**2)
                J = np.sqrt(
                    ds[f"g_11{suffix}"] * ds[f"g_22{suffix}"] - ds[f"g_12{suffix}"] ** 2
                )
            spatial_volume_element = J * dx * dy
        elif xcoord in dims and zcoord in dims:
            # 2d integral on toroidal planes
            # Need to work out area element from metric coefficients. See book by
            # D'haeseleer, Hitchon, Callen and Shohet eq. (2.5.51)
            # J = sqrt(g_11*g_33 - g_13**2)
            J = np.sqrt(
                ds[f"g_11{suffix}"] * ds[f"g_33{suffix}"] - ds[f"g_13{suffix}"] ** 2
            )
            spatial_volume_element = J * dx * dz
        elif ycoord in dims and zcoord in dims:
            # 2d integral on flux-surfaces
            # Need to work out area element from metric coefficients. See book by
            # D'haeseleer, Hitchon, Callen and Shohet eq. (2.5.51)
            # J = sqrt(g_22*g_33 - g_23**2)
            J = np.sqrt(
                ds[f"g_22{suffix}"] * ds[f"g_33{suffix}"] - ds[f"g_23{suffix}"] ** 2
            )
            spatial_volume_element = J * dy * dz
        elif xcoord in dims:
            if variable.direction_y == "Aligned":
                raise ValueError(
                    "Variable is field-aligned, but radial integral along coordinate "
                    "line in globally field-aligned coordinates not supported"
                )
            # 1d radial integral, line element is sqrt(g_11)*dx
            spatial_volume_element = np.sqrt(ds[f"g_11{suffix}"]) * dx
        elif ycoord in dims:
            if variable.direction_y == "Standard":
                # Poloidal integral, line element is e_y projected onto a unit vector in
                # the poloidal direction. e_z is in the toroidal direction and Grad(x)
                # is orthogonal to flux surfaces, so their cross product is in the
                # poloidal direction (within flux surfaces). e_z and Grad(x) are also
                # always orthogonal, so the magnitude of their cross product is the
                # product of their magnitudes. Therefore
                #   e_y.hat{e}_pol = e_y.(e_z x Grad(x))/|Grad(x)||e_z|
                #   e_y.hat{e}_pol = e_y.(e_z x Grad(x))/sqrt(g11*g_33)
                # and using eqs. (2.3.12) and (2.5.22a) from D'haeseleer
                #   e_y.hat{e}_pol = e_y.(e_z x (e_y x e_z / J))/sqrt(g11*g_33)
                #   e_y.hat{e}_pol = e_y.(e_z x (e_y x e_z))/ (J*sqrt(g11*g_33))
                # The double cross product identity is A x (B x C) = (A.C)B - (A.B)C.
                #   e_y.hat{e}_pol = e_y.((e_z.e_z)*e_y - (e_z.e_y)*e_z)/(J*sqrt(g11*g_33))
                #   e_y.hat{e}_pol = e_y.(g_33*e_y - g_23*e_z)/(J*sqrt(g11*g_33))
                #   e_y.hat{e}_pol = (g_33*g_22 - g_23*g_23)/(J*sqrt(g11*g_33))
                # For 'orthogonal' coordinates (radial and poloidal directions are
                # orthogonal) this is equal to 1/sqrt(g22)
                spatial_volume_element = (
                    (
                        ds[f"g_22{suffix}"] * ds[f"g_33{suffix}"]
                        - ds[f"g_23{suffix}"] ** 2
                    )
                    / (
                        ds[f"J{suffix}"]
                        * np.sqrt(ds[f"g11{suffix}"] * ds[f"g_33{suffix}"])
                    )
                    * dy
                )
            elif variable.direction_y == "Aligned":
                # Parallel integral, line element is sqrt(g_22)*dy
                spatial_volume_element = np.sqrt(ds[f"g_22{suffix}"]) * dy
        elif zcoord in dims:
            # Toroidal integral, line element is sqrt(g_33)*dz
            spatial_volume_element = np.sqrt(ds[f"g_33{suffix}"]) * dz
        else:
            # No spatial integral
            spatial_volume_element = 1.0

        spatial_dims = set(dims) - set([tcoord])

        integrand = variable * spatial_volume_element

        # Need to check if the variable being integrated is a Field2D, which does not
        # have a z-dimension to sum over. Other variables are OK because metric
        # coefficients, dx and dy all have both x- and y-dimensions so variable would be
        # broadcast to include them if necessary
        missing_z_sum = zcoord in dims and zcoord not in variable.dims

        # If integrand is a Field2D, need to multiply by nz if integrating over z
        if missing_z_sum:
            spatial_dims -= set(zcoord)
            integral = integrand.sum(dim=spatial_dims)
            integral = integral * ds.sizes[zcoord]
        else:
            integral = integrand.sum(dim=spatial_dims)

        if tcoord in dims:
            if cumulative_t:
                integral = integral.cumulative_integrate(coord=tcoord)
            else:
                integral = integral.integrate(coord=tcoord)

        return integral

    def interpolate_from_unstructured(
        self,
        variables,
        *,
        fill_value=np.nan,
        structured_output=True,
        unstructured_dim_name="unstructured_dim",
        **kwargs,
    ):
        """Interpolate Dataset onto new grids of some existing coordinates

        Parameters
        ----------
        variables : str or sequence of str or ...
            The names of the variables to interpolate. If 'variables=...' is passed
            explicitly, then interpolate all variables in the Dataset.
        **kwargs : (str, array)
            Each keyword is the name of a coordinate in the DataArray, the argument is a
            1d array giving the values of that coordinate on the output grid
        fill_value : float
            fill_value passed through to scipy.interpolation.griddata
        structured_output : bool, default True
            If True, treat output coordinates values as a structured grid.
            If False, output coordinate values must all have the same length and are not
            broadcast together.
        unstructured_dim_name : str, default "unstructured_dim"
            Name used for the dimension in the output that replaces the dimensions of
            the interpolated coordinates. Only used if structured_output=False.

        Returns
        -------
        Dataset
            Dataset interpolated onto a new, structured grid
        """

        if variables is ...:
            variables = [v for v in self.data]
            explicit_variables_arg = False
        else:
            explicit_variables_arg = True

        if isinstance(variables, str):
            variables = [variables]
        if isinstance(variables, tuple):
            variables = list(variables)

        coords_to_interpolate = []
        for coord in self.data.coords:
            if coord not in variables and coord not in kwargs:
                coords_to_interpolate.append(coord)

        ds = xr.Dataset()

        for v in variables + coords_to_interpolate:
            if np.all([c in self.data[v].coords for c in kwargs]):
                ds = ds.merge(
                    self.data[v]
                    .bout.interpolate_from_unstructured(
                        fill_value=fill_value,
                        structured_output=structured_output,
                        unstructured_dim_name=unstructured_dim_name,
                        **kwargs,
                    )
                    .to_dataset()
                )
            elif explicit_variables_arg and v in variables:
                # User explicitly requested v to be interpolated
                raise ValueError(
                    f"Could not interpolate {v} because it does not depend on all "
                    f"coordinates {[c for c in kwargs]}"
                )
            elif v in coords_to_interpolate:
                coords_to_interpolate.remove(v)

        ds = ds.set_coords(coords_to_interpolate)

        return ds

    def interpolate_to_cartesian(
        self, nX=300, nY=300, nZ=100, *, use_float32=True, fill_value=np.nan
    ):
        """
        Interpolate the Dataset to a regular Cartesian grid.

        This method is intended to be used to produce data for visualisation, which
        normally does not require double-precision values, so by default the data is
        converted to `np.float32`. Pass `use_float32=False` to retain the original
        precision.

        Parameters
        ----------
        nX : int (default 300)
            Number of grid points in the X direction
        nY : int (default 300)
            Number of grid points in the Y direction
        nZ : int (default 100)
            Number of grid points in the Z direction
        use_float32 : bool (default True)
            Downgrade precision to `np.float32`?
        fill_value : float (default np.nan)
            Value to use for points outside the interpolation domain (passed to
            `scipy.RegularGridInterpolator`)

        See Also
        --------
        BoutDataArray.interpolate_to_cartesian
        """
        ds = self.data
        ds = ds.bout.add_cartesian_coordinates()

        if not isinstance(use_float32, bool):
            raise ValueError(f"use_float32 must be a bool, got '{use_float32}'")
        if use_float32:
            float_type = np.float32
            ds = ds.astype(float_type)
            for coord in ds.coords:
                # Coordinates are not converted by Dataset.astype, so convert explicitly
                ds[coord] = ds[coord].astype(float_type)
            fill_value = float_type(fill_value)
        else:
            float_type = ds[ds.data_vars[0]].dtype

        tdim = ds.metadata["bout_tdim"]
        zdim = ds.metadata["bout_zdim"]
        if tdim in ds.dims:
            nt = ds.sizes[tdim]
        n_toroidal = ds.sizes[zdim]

        # Create Cartesian grid to interpolate to
        Xmin = ds["X_cartesian"].min()
        Xmax = ds["X_cartesian"].max()
        Ymin = ds["Y_cartesian"].min()
        Ymax = ds["Y_cartesian"].max()
        Zmin = ds["Z_cartesian"].min()
        Zmax = ds["Z_cartesian"].max()
        newX_1d = xr.DataArray(np.linspace(Xmin, Xmax, nX), dims="X")
        newX = newX_1d.expand_dims({"Y": nY, "Z": nZ}, axis=[1, 2])
        newY_1d = xr.DataArray(np.linspace(Ymin, Ymax, nY), dims="Y")
        newY = newY_1d.expand_dims({"X": nX, "Z": nZ}, axis=[0, 2])
        newZ_1d = xr.DataArray(np.linspace(Zmin, Zmax, nZ), dims="Z")
        newZ = newZ_1d.expand_dims({"X": nX, "Y": nY}, axis=[0, 1])
        newR = np.sqrt(newX**2 + newY**2)
        newzeta = np.arctan2(newY, newX)
        # Define newzeta in range 0->2*pi
        newzeta = np.where(newzeta < 0.0, newzeta + 2.0 * np.pi, newzeta)

        from scipy.interpolate import (
            RegularGridInterpolator,
            griddata,
        )

        # Create Cylindrical coordinates for intermediate grid
        Rcyl_min = float_type(ds["R"].min())
        Rcyl_max = float_type(ds["R"].max())
        Zcyl_min = float_type(ds["Z"].min())
        Zcyl_max = float_type(ds["Z"].max())
        n_Rcyl = int(round(nZ * (Rcyl_max - Rcyl_min) / (Zcyl_max - Zcyl_min)))
        Rcyl = xr.DataArray(np.linspace(Rcyl_min, Rcyl_max, 2 * n_Rcyl), dims="r")
        Zcyl = xr.DataArray(np.linspace(Zcyl_min, Zcyl_max, 2 * nZ), dims="z")

        # Create Dataset for result
        result = xr.Dataset()
        result.attrs["metadata"] = ds.metadata

        # Interpolate in two stages for efficiency. Unstructured 3d interpolation is
        # very slow. Unstructured 2d interpolation onto Cartesian (R, Z) grids, followed
        # by structured 3d interpolation onto the (X, Y, Z) grid, is much faster.
        # Structured 3d interpolation straight from (psi, theta, zeta) to (X, Y, Z)
        # leaves artifacts in the output, because theta does not vary continuously
        # everywhere (has branch cuts).

        zeta_out = np.zeros(n_toroidal + 1)
        zeta_out[:-1] = ds[zdim].values
        zeta_out[-1] = zeta_out[-2] + ds["dz"].mean()

        def interp_single_time(da):
            print("    interpolate poloidal planes")

            da_cyl = da.bout.interpolate_from_unstructured(R=Rcyl, Z=Zcyl).transpose(
                "R", "Z", zdim, missing_dims="ignore"
            )

            if zdim not in da_cyl.dims:
                da_cyl = da_cyl.expand_dims({zdim: n_toroidal + 1}, axis=-1)
            else:
                # Impose toroidal periodicity by appending zdim=0 to end of array
                da_cyl = xr.concat((da_cyl, da_cyl.isel({zdim: 0})), zdim)

            print("    build 3d interpolator")
            interp = RegularGridInterpolator(
                (Rcyl.values, Zcyl.values, zeta_out),
                da_cyl.values,
                bounds_error=False,
                fill_value=fill_value,
            )

            print("    do 3d interpolation")
            return interp(
                (newR, newZ, newzeta),
                method="linear",
            )

        for name, da in ds.data_vars.items():
            print(f"\ninterpolating {name}")
            # order of dimensions does not really matter here - output only depends on
            # shape of newR, newZ, newzeta. Possibly more efficient to assign the 2d
            # results in the loop to the last two dimensions, so put zeta first.  Can't
            # just use da.min().item() here (to get a scalar value instead of a
            # zero-size array) because .item() doesn't work for dask arrays (yet!).

            datamin = float_type(da.min().values)
            datamax = float_type(da.max().values)

            if tdim in da.dims:
                data_cartesian = np.zeros((nt, nX, nY, nZ), dtype=float_type)
                for tind in range(nt):
                    print(f"  tind={tind}")
                    data_cartesian[tind, :, :, :] = interp_single_time(
                        da.isel({tdim: tind})
                    )
                result[name] = xr.DataArray(data_cartesian, dims=[tdim, "X", "Y", "Z"])
            else:
                data_cartesian = interp_single_time(da)
                result[name] = xr.DataArray(data_cartesian, dims=["X", "Y", "Z"])

            # Copy metadata to data variables, in case it is needed
            result[name].attrs["metadata"] = ds.metadata

        result = result.assign_coords(X=newX_1d, Y=newY_1d, Z=newZ_1d)

        return result

    def add_cartesian_coordinates(self):
        """
        Add Cartesian (X,Y,Z) coordinates.

        Returns
        -------
        Dataset with new coordinates added, which are named 'X_cartesian',
        'Y_cartesian', and 'Z_cartesian'
        """
        return _add_cartesian_coordinates(self.data)

    def remove_yboundaries(self, **kwargs):
        """
        Remove y-boundary points, if present, from the Dataset
        """

        variables = []
        xcoord = self.data.metadata["bout_xdim"]
        ycoord = self.data.metadata["bout_ydim"]
        new_metadata = None
        for v in self.data:
            if xcoord in self.data[v].dims and ycoord in self.data[v].dims:
                variables.append(
                    self.data[v].bout.remove_yboundaries(return_dataset=True, **kwargs)
                )
                new_metadata = variables[-1].metadata
            elif ycoord in self.data[v].dims:
                raise ValueError(
                    f"{v} only has a {ycoord}-dimension so cannot split "
                    f"into regions."
                )
            else:
                variable = self.data[v]
                if "keep_yboundaries" in variable.metadata:
                    variable.attrs["metadata"] = copy(variable.metadata)
                    variable.metadata["keep_yboundaries"] = 0
                variables.append(variable.bout.to_dataset())
        if new_metadata is None:
            # were no 2d or 3d variables so do not have updated jyseps*, ny_inner but
            # does not matter because missing metadata is only useful for 2d or 3d
            # variables
            new_metadata = variables[0].metadata

        result = xr.merge(variables)

        result.attrs = copy(self.data.attrs)

        # Copy metadata to get possibly modified jyseps*, ny_inner, ny
        result.attrs["metadata"] = new_metadata

        if "regions" in result.attrs:
            # regions are not correct for modified BoutDataset
            del result.attrs["regions"]

        # call to re-create regions
        result = apply_geometry(result, self.data.geometry)

        return result

    def get_bounding_surfaces(self, coords=("R", "Z")):
        """
        Get bounding surfaces.
        Surfaces are returned as arrays of points describing a polygon, assuming the
        third spatial dimension is a symmetry direction.

        Parameters
        ----------
        coords : (str, str), default ("R", "Z")
            Pair of names of coordinates whose values are used to give the positions of
            the points in the result

        Returns
        -------
        result : list of DataArrays
            Each DataArray in the list contains points on a boundary, with size
            (<number of points in the bounding polygon>, 2). Points wind clockwise around
            the outside domain, and anti-clockwise around the inside (if there is an
            inner boundary).
        """
        return _get_bounding_surfaces(self.data, coords)

    def save(
        self,
        savepath="./boutdata.nc",
        filetype="NETCDF4",
        variables=None,
        save_dtype=None,
        separate_vars=False,
        pre_load=False,
    ):
        """
        Save data variables to a netCDF file.

        Parameters
        ----------
        savepath : str, optional
        filetype : str, optional
        variables : list of str, optional
            Variables from the dataset to save. Default is to save all of them.
        separate_vars: bool, optional
            If this is true then every variable which depends on time (but not
            solely on time) will be saved into a different output file.
            The files are labelled by the name of the variable. Variables which
            don't meet this criterion will be present in every output file.
        pre_load : bool, optional
            When saving separate variables, will load each variable into memory
            before saving to file, which can be considerably faster.

        Examples
        --------
        If `separate_vars=True`, then multiple files will be created. These can
        all be opened and merged in one go using a call of the form:

        ds = xr.open_mfdataset('boutdata_*.nc', combine='nested', concat_dim=None)
        """

        if variables is None:
            # Save all variables
            to_save = self.data
        else:
            to_save = self.data[variables]

        if savepath == "./boutdata.nc":
            print(
                "Will save data into the current working directory, named as"
                " boutdata_[var].nc"
            )
        if savepath is None:
            raise ValueError("Must provide a path to which to save the data.")

        # make shallow copy of Dataset, so we do not modify the attributes of the data
        # when we change things to save
        to_save = to_save.copy()

        options = to_save.attrs.pop("options")
        if options:
            # TODO Convert Ben's options class to a (flattened) nested
            # dictionary then store it in ds.attrs?
            warnings.warn(
                "Haven't decided how to write options file back out yet - deleting "
                "options for now. To re-load this Dataset, pass the same inputfilepath "
                "to open_boutdataset when re-loading."
            )
        # Delete placeholders for options on each variable and coordinate
        for var in chain(to_save.data_vars, to_save.coords):
            try:
                del to_save[var].attrs["options"]
            except KeyError:
                pass

        # Store the metadata as individual attributes instead because
        # netCDF can't handle storing arbitrary objects in attrs
        def dict_to_attrs(obj, section):
            for key, value in obj.attrs.pop(section).items():
                obj.attrs[section + ":" + key] = value

        dict_to_attrs(to_save, "metadata")
        # Must do this for all variables and coordinates in dataset too
        for varname, da in chain(to_save.data_vars.items(), to_save.coords.items()):
            try:
                dict_to_attrs(da, "metadata")
            except KeyError:
                pass

        if "regions" in to_save.attrs:
            # Do not need to save regions as these can be reconstructed from the metadata
            try:
                del to_save.attrs["regions"]
            except KeyError:
                pass
            for var in chain(to_save.data_vars, to_save.coords):
                try:
                    del to_save[var].attrs["regions"]
                except KeyError:
                    pass

        if save_dtype is not None:
            encoding = {v: {"dtype": save_dtype} for v in to_save}
        else:
            encoding = None

        if separate_vars:
            # Save each major variable to a different netCDF file

            # Determine which variables are "major"
            # Defined as time-dependent, but not solely time-dependent
            major_vars, minor_vars = _find_major_vars(to_save)

            print("Will save the variables {} separately".format(str(major_vars)))

            # Save each one to separate file
            # TODO perform the save in parallel with save_mfdataset?
            for major_var in major_vars:
                # Group variables so that there is only one time-dependent
                # variable saved in each file
                minor_data = [to_save[minor_var] for minor_var in minor_vars]
                single_var_ds = xr.merge([to_save[major_var], *minor_data])

                # Add the attrs back on
                single_var_ds.attrs = to_save.attrs

                if pre_load:
                    single_var_ds.load()

                # Include the name of the variable in the name of the saved
                # file
                path = Path(savepath)
                var_savepath = (
                    str(path.parent / path.stem) + "_" + str(major_var) + path.suffix
                )
                if encoding is not None:
                    var_encoding = {major_var: encoding[major_var]}
                else:
                    var_encoding = None
                print("Saving " + major_var + " data...")
                with ProgressBar():
                    single_var_ds.to_netcdf(
                        path=str(var_savepath),
                        format=filetype,
                        compute=True,
                        encoding=var_encoding,
                    )

                # Force memory deallocation to limit RAM usage
                single_var_ds.close()
                del single_var_ds
                gc.collect()
        else:
            # Save data to a single file
            print("Saving data...")
            with ProgressBar():
                to_save.to_netcdf(
                    path=savepath, format=filetype, compute=True, encoding=encoding
                )

        return

    def to_restart(
        self,
        variables=None,
        *,
        savepath=".",
        nxpe=None,
        nype=None,
        tind=-1,
        prefix="BOUT.restart",
        overwrite=False,
    ):
        """
        Write out a timestep as a set of netCDF BOUT.restart files.

        If processor decomposition is not specified then data will be saved
        using the decomposition it had when loaded.

        Parameters
        ----------
        variables : str or sequence of str, optional
            The evolving variables needed in the restart files. If not given explicitly,
            all time-evolving variables in the Dataset will be used, which may result in
            larger restart files than necessary. If there is no time-dimension in the
            Dataset (e.g. if it was loaded from restart files), then all variables will
            be added if this argument is not given explicitly.
        savepath : str, default '.'
            Directory to save the created restart files under
        nxpe : int, optional
            Number of processors in the x-direction. If not given, keep the number used
            for the original simulation
        nype : int, optional
            Number of processors in the y-direction. If not given, keep the number used
            for the original simulation
        tind : int, default -1
            Time-index of the slice to write to the restart files. Note, when creating
            restart files from 'dump' files it is recommended to open the Dataset using
            the full time range and use the `tind` argument here, rather than selecting
            a time point manually, so that the calculation of `hist_hi` in the output
            can be correct (which requires knowing the existing value of `hist_hi`
            (output step count at the end of the simulation), `tind` and the total
            number of time points in the current output data).
        prefix : str, default "BOUT.restart"
            Prefix to use for names of restart files
        overwrite : bool, default False
            By default, raises if restart file already exists. Set to True to overwrite
            existing files
        """

        if isinstance(variables, str):
            variables = [variables]

        # Set processor decomposition if not given
        if nxpe is None:
            nxpe = self.metadata["NXPE"]
        if nype is None:
            nype = self.metadata["NYPE"]

        # Is this even possible without saving the guard cells?
        # Can they be recreated?
        restart_datasets, paths = _split_into_restarts(
            self.data,
            variables,
            savepath,
            nxpe,
            nype,
            tind,
            prefix,
            overwrite,
        )

        with ProgressBar():
            xr.save_mfdataset(restart_datasets, paths, compute=True)

    def animate_list(
        self,
        variables,
        animate_over=None,
        save_as=None,
        show=False,
        fps=10,
        nrows=None,
        ncols=None,
        poloidal_plot=False,
        axis_coords=None,
        subplots_adjust=None,
        vmin=None,
        vmax=None,
        logscale=None,
        titles=None,
        aspect=None,
        extend=None,
        controls="both",
        tight_layout=True,
        **kwargs,
    ):
        """
        Parameters
        ----------
        variables : list of str or BoutDataArray
            The variables to plot. For any string passed, the corresponding
            variable in this DataSet is used - then the calling DataSet must
            have only 3 dimensions. It is possible to pass BoutDataArrays to
            allow more flexible plots, e.g. with different variables being
            plotted against different axes.
        animate_over : str, optional
            Dimension over which to animate, defaults to the time dimension
        save_as : str, optional
            If passed, a gif is created with this filename
        show : bool, optional
            Call pyplot.show() to display the animation
        fps : float, optional
            Indicates the number of frames per second to play
        nrows : int, optional
            Specify the number of rows of plots
        ncols : int, optional
            Specify the number of columns of plots
        poloidal_plot : bool or sequence of bool, optional
            If set to True, make all 2D animations in the poloidal plane instead of using
            grid coordinates, per variable if sequence is given
        axis_coords : None, str, dict or list of None, str or dict
            Coordinates to use for axis labelling.
            - None: Use the dimension coordinate for each axis, if it exists.
            - "index": Use the integer index values.
            - dict: keys are dimension names, values set axis_coords for each axis
              separately. Values can be: None, "index", the name of a 1d variable or
              coordinate (which must have the dimension given by 'key'), or a 1d
              numpy array, dask array or DataArray whose length matches the length of
              the dimension given by 'key'.
            Only affects time coordinate for plots with poloidal_plot=True.
            If a list is passed, it must have the same length as 'variables' and gives
            the axis_coords setting for each plot individually.
            The setting to use for the 'animate_over' coordinate can be passed in one or
            more dict values, but must be the same in all dicts if given more than once.
        subplots_adjust : dict, optional
            Arguments passed to fig.subplots_adjust()()
        vmin : float or sequence of floats
            Minimum value for color scale, per variable if a sequence is given
        vmax : float or sequence of floats
            Maximum value for color scale, per variable if a sequence is given
        logscale : bool or float, sequence of bool or float, optional
            If True, default to a logarithmic color scale instead of a linear one.
            If a non-bool type is passed it is treated as a float used to set the linear
            threshold of a symmetric logarithmic scale as
            linthresh=min(abs(vmin),abs(vmax))*logscale, defaults to 1e-5 if True is
            passed.
            Per variable if sequence is given.
        titles : sequence of str or None, optional
            Custom titles for each plot. Pass None in the sequence to use the default for
            a certain variable
        aspect : str or None, or sequence of str or None, optional
            Argument to set_aspect() for each plot. Defaults to "equal" for poloidal
            plots and "auto" for others.
        extend : str or None, optional
            Passed to fig.colorbar()
        controls : string or None, default "both"
            By default, add both the timeline and play/pause toggle to the animation. If
            "timeline" is passed add only the timeline, if "toggle" is passed add only
            the play/pause toggle. If None or an empty string is passed, add neither.
        tight_layout : bool or dict, optional
            If set to False, don't call tight_layout() on the figure.
            If a dict is passed, the dict entries are passed as arguments to
            tight_layout()
        **kwargs : dict, optional
            Additional keyword arguments are passed on to each animation function, per
            variable if a sequence is given.

        Returns
        -------
        animation
            An animatplot.Animation object.
        """

        if animate_over is None:
            animate_over = self.metadata.get("bout_tdim", "t")

        nvars = len(variables)

        if nrows is None and ncols is None:
            ncols = int(np.ceil(np.sqrt(nvars)))
            nrows = int(np.ceil(nvars / ncols))
        elif nrows is None:
            nrows = int(np.ceil(nvars / ncols))
        elif ncols is None:
            ncols = int(np.ceil(nvars / nrows))
        else:
            if nrows * ncols < nvars:
                raise ValueError("Not enough rows*columns to fit all variables")

        fig, axes = plt.subplots(nrows, ncols, squeeze=False)
        axes = axes.flatten()

        ncells = nrows * ncols

        if nvars < ncells:
            for index in range(ncells - nvars):
                fig.delaxes(axes[ncells - index - 1])

        if subplots_adjust is not None:
            fig.subplots_adjust(**subplots_adjust)

        def _expand_list_arg(arg, arg_name):
            if isinstance(arg, collections.abc.Sequence) and not isinstance(arg, str):
                if len(arg) != len(variables):
                    raise ValueError(
                        "if %s is a sequence, it must have the same "
                        'number of elements as "variables"' % arg_name
                    )
            else:
                arg = [arg] * len(variables)
            return arg

        poloidal_plot = _expand_list_arg(poloidal_plot, "poloidal_plot")
        vmin = _expand_list_arg(vmin, "vmin")
        vmax = _expand_list_arg(vmax, "vmax")
        logscale = _expand_list_arg(logscale, "logscale")
        titles = _expand_list_arg(titles, "titles")
        aspect = _expand_list_arg(aspect, "aspect")
        extend = _expand_list_arg(extend, "extend")
        axis_coords = _expand_list_arg(axis_coords, "axis_coords")
        for k in kwargs:
            kwargs[k] = _expand_list_arg(kwargs[k], k)

        blocks = []

        def is_list(variable):
            return (
                isinstance(variable, list)
                or isinstance(variable, tuple)
                or isinstance(variable, set)
            )

        for i, subplot_args in enumerate(
            zip(
                variables,
                axes,
                poloidal_plot,
                axis_coords,
                vmin,
                vmax,
                logscale,
                titles,
                aspect,
                extend,
            )
        ):

            (
                v,
                ax,
                this_poloidal_plot,
                this_axis_coords,
                this_vmin,
                this_vmax,
                this_logscale,
                this_title,
                this_aspect,
                this_extend,
            ) = subplot_args

            this_kwargs = {k: v[i] for k, v in kwargs.items()}

            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.1)

            if is_list(v):
                for i in range(len(v)):
                    if isinstance(v[i], str):
                        v[i] = self.data[v[i]]
                # list of variables for one subplot only supported for line plots with 1
                # dimension plus time
                ndims = 2
                dims = v[0].dims
                if len(dims) != 2:
                    raise ValueError(
                        "Variables in sublist must be 2d - can only overlay line plots"
                    )
                for w in v:
                    if not w.dims == dims:
                        raise ValueError(
                            f"All variables in sub-list must have same dimensions."
                            f"{v[0].name} had {v[0].dims} but {w.name} had {w.dims}."
                        )
            else:
                if isinstance(v, str):
                    v = self.data[v]

                data = v.bout.data
                ndims = len(data.dims)
                ax.set_title(data.name)

            if ndims == 2:
                if not is_list(v):
                    blocks.append(
                        animate_line(
                            data=data,
                            ax=ax,
                            animate_over=animate_over,
                            animate=False,
                            axis_coords=this_axis_coords,
                            aspect=this_aspect,
                            vmin=this_vmin,
                            vmax=this_vmax,
                            **this_kwargs,
                        )
                    )
                else:
                    for w in v:
                        blocks.append(
                            animate_line(
                                data=w,
                                ax=ax,
                                animate_over=animate_over,
                                animate=False,
                                axis_coords=this_axis_coords,
                                aspect=this_aspect,
                                vmin=this_vmin,
                                vmax=this_vmax,
                                label=w.name,
                                **this_kwargs,
                            )
                        )
                    legend = ax.legend()
                    legend.set_draggable(True)
                    # set 'v' to use for the timeline below
                    v = v[0]
            elif ndims == 3:
                if this_poloidal_plot:
                    var_blocks = animate_poloidal(
                        data,
                        ax=ax,
                        cax=cax,
                        animate_over=animate_over,
                        animate=False,
                        axis_coords=this_axis_coords,
                        vmin=this_vmin,
                        vmax=this_vmax,
                        logscale=this_logscale,
                        aspect=this_aspect,
                        extend=this_extend,
                        **this_kwargs,
                    )
                    for block in var_blocks:
                        blocks.append(block)
                else:
                    blocks.append(
                        animate_pcolormesh(
                            data=data,
                            ax=ax,
                            cax=cax,
                            animate_over=animate_over,
                            animate=False,
                            axis_coords=this_axis_coords,
                            vmin=this_vmin,
                            vmax=this_vmax,
                            logscale=this_logscale,
                            aspect=this_aspect,
                            extend=this_extend,
                            **this_kwargs,
                        )
                    )
            else:
                raise ValueError(
                    "Unsupported number of dimensions "
                    + str(ndims)
                    + ". Dims are "
                    + str(v.dims)
                )

            if this_title is not None:
                # Replace default title with user-specified one
                ax.set_title(this_title)

        if np.all([a == "index" for a in axis_coords]):
            time_opt = "index"
        elif np.any([isinstance(a, dict) and animate_over in a for a in axis_coords]):
            given_values = [
                a[animate_over]
                for a in axis_coords
                if isinstance(a, dict) and animate_over in a
            ]
            time_opt = given_values[0]
            if len(given_values) > 1 and not np.all(
                [v == time_opt for v in given_values[1:]]
            ):
                raise ValueError(
                    f"Inconsistent axis_coords values given for animate_over "
                    f"coordinate ({animate_over}). Got {given_values}."
                )
        else:
            time_opt = None
        time_values, time_label = _parse_coord_option(animate_over, time_opt, self.data)
        time_values, time_suffix = _normalise_time_coord(time_values)

        timeline = amp.Timeline(time_values, fps=fps, units=time_suffix)
        anim = amp.Animation(blocks, timeline)

        if tight_layout:
            if subplots_adjust is not None:
                warnings.warn(
                    "tight_layout argument to animate_list() is True, but "
                    "subplots_adjust argument is not None. subplots_adjust "
                    "is being ignored."
                )
            if not isinstance(tight_layout, dict):
                tight_layout = {}
            fig.tight_layout(**tight_layout)

        _add_controls(anim, controls, time_label)

        if save_as is not None:
            anim.save(save_as + ".gif", writer=PillowWriter(fps=fps))

        if show:
            plt.show()

        return anim


def _find_major_vars(data):
    """
    Splits data into those variables likely to require a lot of storage space
    (defined as those which depend on time and at least one other dimension).
    These are normally the variables of physical interest.
    """

    # TODO Use an Ordered Set instead to preserve order of variables in files?
    tcoord = data.attrs.get("metadata:bout_tdim", "t")
    major_vars = set(
        var
        for var in data.data_vars
        if (tcoord in data[var].dims) and data[var].dims != (tcoord,)
    )
    minor_vars = set(data.data_vars) - set(major_vars)
    return list(major_vars), list(minor_vars)



from collections import namedtuple
from copy import deepcopy
import inspect
from pathlib import Path
import re
from functools import reduce
import operator

import pytest

import numpy as np

from xarray import DataArray, Dataset, concat
from xarray.tests.test_dataset import create_test_data
import xarray.testing as xrt

from natsort import natsorted

from xbout.load import (
    _check_filetype,
    _expand_wildcards,
    _expand_filepaths,
    _arrange_for_concatenation,
    _trim,
    _infer_contains_boundaries,
    open_boutdataset,
    _BOUT_PER_PROC_VARIABLES,
    _BOUT_TIME_DEPENDENT_META_VARS,
)
from xbout.utils import _separate_metadata
from xbout.tests.utils_for_tests import _get_kwargs


def test_check_extensions(tmp_path):
    files_dir = tmp_path.joinpath("data")
    files_dir.mkdir()
    example_nc_file = files_dir.joinpath("example.nc")
    example_nc_file.write_text("content_nc")

    filetype = _check_filetype(example_nc_file)
    assert filetype == "netcdf4"

    example_hdf5_file = files_dir.joinpath("example.h5netcdf")
    example_hdf5_file.write_text("content_hdf5")

    filetype = _check_filetype(example_hdf5_file)
    assert filetype == "h5netcdf"

    example_invalid_file = files_dir.joinpath("example.txt")
    example_hdf5_file.write_text("content_txt")
    with pytest.raises(IOError):
        filetype = _check_filetype(example_invalid_file)


def test_set_fci_coords(create_example_grid_file_fci, create_example_files_fci):
    grid = create_example_grid_file_fci
    data = create_example_files_fci

    ds = open_boutdataset(data, gridfilepath=grid, geometry="fci")
    assert "R" in ds
    assert "Z" in ds


class TestPathHandling:
    def test_glob_expansion_single(self, tmp_path):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()
        example_file = files_dir.joinpath("example.0.nc")
        example_file.write_text("content")

        path = example_file
        filepaths = _expand_wildcards(path)
        assert filepaths[0] == example_file

        path = files_dir.joinpath("example.*.nc")
        filepaths = _expand_wildcards(path)
        assert filepaths[0] == example_file

    @pytest.mark.parametrize(
        "ii, jj", [(1, 1), (1, 4), (3, 1), (5, 3), (12, 1), (1, 12), (121, 2), (3, 111)]
    )
    def test_glob_expansion_both(self, tmp_path, ii, jj):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()
        filepaths = []
        for i in range(ii):
            example_run_dir = files_dir.joinpath("run" + str(i))
            example_run_dir.mkdir()
            for j in range(jj):
                example_file = example_run_dir.joinpath("example." + str(j) + ".nc")
                example_file.write_text("content")
                filepaths.append(example_file)
        expected_filepaths = natsorted(filepaths, key=lambda filepath: str(filepath))

        path = files_dir.joinpath("run*/example.*.nc")
        actual_filepaths = _expand_wildcards(path)

        assert actual_filepaths == expected_filepaths

    @pytest.mark.parametrize(
        "ii, jj", [(1, 1), (1, 4), (3, 1), (5, 3), (1, 12), (3, 111)]
    )
    def test_glob_expansion_brackets(self, tmp_path, ii, jj):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()
        filepaths = []
        for i in range(ii):
            example_run_dir = files_dir.joinpath("run" + str(i))
            example_run_dir.mkdir()
            for j in range(jj):
                example_file = example_run_dir.joinpath("example." + str(j) + ".nc")
                example_file.write_text("content")
                filepaths.append(example_file)
        expected_filepaths = natsorted(filepaths, key=lambda filepath: str(filepath))

        path = files_dir.joinpath("run[1-9]/example.*.nc")
        actual_filepaths = _expand_wildcards(path)

        assert actual_filepaths == expected_filepaths[jj:]

    def test_no_files(self, tmp_path):
        files_dir = tmp_path.joinpath("data")
        files_dir.mkdir()

        with pytest.raises(IOError):
            path = files_dir.joinpath("run*/example.*.nc")
            actual_filepaths = _expand_filepaths(path)


@pytest.fixture()
def create_filepaths():
    return _create_filepaths


def _create_filepaths(nxpe=1, nype=1, nt=1):
    filepaths = []
    for t in range(nt):
        for i in range(nype):
            for j in range(nxpe):
                file_num = j + nxpe * i
                path = "./run{}".format(str(t)) + "/BOUT.dmp.{}.nc".format(
                    str(file_num)
                )
                filepaths.append(path)

    return filepaths


class TestArrange:
    def test_arrange_single(self, create_filepaths):
        paths = create_filepaths(nxpe=1, nype=1, nt=1)
        expected_path_grid = [[["./run0/BOUT.dmp.0.nc"]]]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=1, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, None, None]

    def test_arrange_along_x(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=1, nt=1)
        expected_path_grid = [
            [["./run0/BOUT.dmp.0.nc", "./run0/BOUT.dmp.1.nc", "./run0/BOUT.dmp.2.nc"]]
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, None, "x"]

    def test_arrange_along_y(self, create_filepaths):
        paths = create_filepaths(nxpe=1, nype=3, nt=1)
        expected_path_grid = [
            [
                ["./run0/BOUT.dmp.0.nc"],
                ["./run0/BOUT.dmp.1.nc"],
                ["./run0/BOUT.dmp.2.nc"],
            ]
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=1, nype=3
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, "y", None]

    def test_arrange_along_t(self, create_filepaths):
        paths = create_filepaths(nxpe=1, nype=1, nt=3)
        expected_path_grid = [
            [["./run0/BOUT.dmp.0.nc"]],
            [["./run1/BOUT.dmp.0.nc"]],
            [["./run2/BOUT.dmp.0.nc"]],
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=1, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == ["t", None, None]

    def test_arrange_along_xy(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=2, nt=1)
        expected_path_grid = [
            [
                [
                    "./run0/BOUT.dmp.0.nc",
                    "./run0/BOUT.dmp.1.nc",
                    "./run0/BOUT.dmp.2.nc",
                ],
                [
                    "./run0/BOUT.dmp.3.nc",
                    "./run0/BOUT.dmp.4.nc",
                    "./run0/BOUT.dmp.5.nc",
                ],
            ]
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=2
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == [None, "y", "x"]

    def test_arrange_along_xt(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=1, nt=2)
        expected_path_grid = [
            [["./run0/BOUT.dmp.0.nc", "./run0/BOUT.dmp.1.nc", "./run0/BOUT.dmp.2.nc"]],
            [["./run1/BOUT.dmp.0.nc", "./run1/BOUT.dmp.1.nc", "./run1/BOUT.dmp.2.nc"]],
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=1
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == ["t", None, "x"]

    def test_arrange_along_xyt(self, create_filepaths):
        paths = create_filepaths(nxpe=3, nype=2, nt=2)
        expected_path_grid = [
            [
                [
                    "./run0/BOUT.dmp.0.nc",
                    "./run0/BOUT.dmp.1.nc",
                    "./run0/BOUT.dmp.2.nc",
                ],
                [
                    "./run0/BOUT.dmp.3.nc",
                    "./run0/BOUT.dmp.4.nc",
                    "./run0/BOUT.dmp.5.nc",
                ],
            ],
            [
                [
                    "./run1/BOUT.dmp.0.nc",
                    "./run1/BOUT.dmp.1.nc",
                    "./run1/BOUT.dmp.2.nc",
                ],
                [
                    "./run1/BOUT.dmp.3.nc",
                    "./run1/BOUT.dmp.4.nc",
                    "./run1/BOUT.dmp.5.nc",
                ],
            ],
        ]
        actual_path_grid, actual_concat_dims = _arrange_for_concatenation(
            paths, nxpe=3, nype=2
        )
        assert expected_path_grid == actual_path_grid
        assert actual_concat_dims == ["t", "y", "x"]


@pytest.fixture(scope="session")
def bout_xyt_example_files(tmp_path_factory):
    return _bout_xyt_example_files


_bout_xyt_example_files_cache = {}


def _bout_xyt_example_files(
    tmp_path_factory,
    prefix="BOUT.dmp",
    lengths=(6, 2, 4, 7),
    nxpe=4,
    nype=2,
    nt=1,
    guards=None,
    syn_data_type="random",
    grid=None,
    squashed=False,
    topology="core",
    write_to_disk=False,
    bout_v5=False,
    metric_3D=False,
):
    """
    Mocks up a set of BOUT-like Datasets

    Either returns list of Datasets (if write_to_disk=False)
    or writes Datasets to netCDF files and returns the temporary test directory
    containing them, deleting the temporary directory once that test is done (if
    write_to_disk=True).
    """
    call_args = _get_kwargs(ignore="tmp_path_factory")

    try:
        # Has been called with the same signature before, just return the cached result
        return deepcopy(_bout_xyt_example_files_cache[call_args])
    except KeyError:
        pass

    if guards is None:
        guards = {}

    mxg = guards.get("x", 0)
    myg = guards.get("y", 0)

    if squashed:
        # create a single data-file, but alter the 'nxpe' and 'nype' variables, as if the
        # file had been created by combining a set of BOUT.dmp.*.nc files
        this_lengths = (
            lengths[0],
            lengths[1] * nxpe,
            lengths[2] * nype,
            lengths[3],
        )
        ds_list, file_list = create_bout_ds_list(
            prefix=prefix,
            lengths=this_lengths,
            nxpe=1,
            nype=1,
            nt=nt,
            guards=guards,
            topology=topology,
            syn_data_type=syn_data_type,
            squashed=True,
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )
        ds_list[0]["nxpe"] = nxpe
        ds_list[0]["nype"] = nype
    else:
        ds_list, file_list = create_bout_ds_list(
            prefix=prefix,
            lengths=lengths,
            nxpe=nxpe,
            nype=nype,
            nt=nt,
            guards=guards,
            topology=topology,
            syn_data_type=syn_data_type,
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )

    if grid is not None:
        xsize = lengths[1] * nxpe
        ysize = lengths[2] * nype
        grid_ds = create_bout_grid_ds(
            xsize=xsize,
            ysize=ysize,
            guards=guards,
            topology=topology,
            ny_inner=3 * lengths[2],
        )

    if not write_to_disk:
        if grid is None:
            _bout_xyt_example_files_cache[call_args] = ds_list
            return deepcopy(ds_list)
        else:
            _bout_xyt_example_files_cache[call_args] = ds_list, grid_ds
            return deepcopy((ds_list, grid_ds))
        raise ValueError("tmp_path_factory required when write_to_disk=True")

    save_dir = tmp_path_factory.mktemp("data")

    for ds, file_name in zip(ds_list, file_list):
        ds.to_netcdf(save_dir.joinpath(file_name))

    if grid is not None:
        grid_ds.to_netcdf(save_dir.joinpath(grid + ".nc"))

    # Return a glob-like path to all files created, which has all file numbers replaced
    # with a single asterix
    path = str(save_dir.joinpath(file_list[-1]))

    count = 1
    if nt > 1:
        count += 1
    # We have to reverse the path before limiting the number of numbers replaced so that the
    # tests don't get confused by pytest's persistent temporary directories (which are also designated
    # by different numbers)
    glob_pattern = Path((re.sub(r"\d+", "*", path[::-1], count=count))[::-1])
    _bout_xyt_example_files_cache[call_args] = glob_pattern
    return glob_pattern


def create_bout_ds_list(
    prefix,
    lengths=(6, 2, 4, 7),
    nxpe=4,
    nype=2,
    nt=1,
    guards={},
    topology="core",
    syn_data_type="random",
    squashed=False,
    bout_v5=False,
    metric_3D=False,
):
    """
    Mocks up a set of BOUT-like datasets.

    Structured as though they were produced by a x-y parallelised run with multiple restarts.
    """

    if nt != 1:
        raise ValueError(
            "nt > 1 means the time dimension is split over several "
            + "directories. This is not implemented yet."
        )

    file_list = []
    ds_list = []
    for j in range(nype):
        for i in range(nxpe):
            num = i + nxpe * j
            filename = prefix + "." + str(num) + ".nc"
            file_list.append(filename)

            # Include guard cells
            upper_bndry_cells = {dim: guards.get(dim) for dim in guards.keys()}
            lower_bndry_cells = {dim: guards.get(dim) for dim in guards.keys()}

            ds = create_bout_ds(
                syn_data_type=syn_data_type,
                num=num,
                lengths=lengths,
                nxpe=nxpe,
                nype=nype,
                xproc=i,
                yproc=j,
                guards=guards,
                topology=topology,
                squashed=squashed,
                bout_v5=bout_v5,
                metric_3D=metric_3D,
            )
            ds_list.append(ds)

    return ds_list, file_list


_create_bout_ds_cache = {}


def create_bout_ds(
    syn_data_type="random",
    lengths=(6, 2, 4, 7),
    num=0,
    nxpe=1,
    nype=1,
    xproc=0,
    yproc=0,
    guards=None,
    topology="core",
    squashed=False,
    bout_v5=False,
    metric_3D=False,
):
    call_args = _get_kwargs()

    try:
        # Has been called with the same signature before, just return the cached result
        return deepcopy(_create_bout_ds_cache[call_args])
    except KeyError:
        pass

    if metric_3D and not bout_v5:
        raise ValueError("3D metric requires BOUT++ v5")

    if guards is None:
        guards = {}

    # Set the shape of the data in this dataset
    t_length, x_length, y_length, z_length = lengths
    mxg = guards.get("x", 0)
    myg = guards.get("y", 0)
    x_length += 2 * mxg
    y_length += 2 * myg

    # calculate global nx, ny and nz
    nx = nxpe * lengths[1] + 2 * mxg
    ny = nype * lengths[2]
    nz = 1 * lengths[3]

    if squashed and "double-null" in topology:
        ny = ny + 2 * myg
        y_length = y_length + 2 * myg
    shape = (t_length, x_length, y_length, z_length)

    # Fill with some kind of synthetic data
    if syn_data_type == "random":
        # Each dataset contains unique random noise
        np.random.seed(seed=num)
        data = np.random.randn(*shape)
    elif syn_data_type == "linear":
        # Variables increase linearly across entire domain
        data = DataArray(-np.ones(shape), dims=("t", "x", "y", "z"))

        t_array = DataArray(
            (nx - 2 * mxg) * ny * nz * np.arange(t_length, dtype=float), dims="t"
        )
        x_array = DataArray(
            ny * nz * (xproc * lengths[1] + np.arange(lengths[1], dtype=float)),
            dims="x",
        )
        y_array = DataArray(
            nz * (yproc * lengths[2] + np.arange(lengths[2], dtype=float)), dims="y"
        )
        z_array = DataArray(np.arange(z_length, dtype=float), dims="z")

        data[:, mxg : x_length - mxg, myg : lengths[2] + myg, :] = (
            t_array + x_array + y_array + z_array
        )
    elif syn_data_type == "stepped":
        # Each dataset contains a different number depending on the filename
        data = np.ones(shape) * num
    elif isinstance(syn_data_type, int):
        data = np.ones(shape) * syn_data_type
    else:
        raise ValueError("Not a recognised choice of type of synthetic bout data.")

    T = DataArray(data, dims=["t", "x", "y", "z"])
    n = DataArray(data, dims=["t", "x", "y", "z"])
    S = DataArray(data[:, :, :, 0], dims=["t", "x", "y"])
    for v in [n, T]:
        v.attrs["direction_y"] = "Standard"
        v.attrs["cell_location"] = "CELL_CENTRE"
        v.attrs["direction_z"] = "Standard"
    for v in [S]:
        v.attrs["direction_y"] = "Standard"
        v.attrs["cell_location"] = "CELL_CENTRE"
        v.attrs["direction_z"] = "Average"
    ds = Dataset({"n": n, "T": T, "S": S})

    # BOUT_VERSION needed to deal with backwards incompatible changes:
    #
    # - v3 and earlier: number of points in z is MZ-1
    # - v4 and later: number of points in z is MZ
    # - v5 and later: metric components can be either 2D or 3D
    # - v5 and later: dz changed to be a Field2D/3D
    ds["BOUT_VERSION"] = 5.0 if bout_v5 else 4.3
    ds["use_metric_3d"] = int(metric_3D)

    # Include grid data
    ds["NXPE"] = nxpe
    ds["NYPE"] = nype
    ds["NZPE"] = 1
    ds["PE_XIND"] = xproc
    ds["PE_YIND"] = yproc
    ds["MYPE"] = num

    ds["MXG"] = mxg
    ds["MYG"] = myg
    ds["MZG"] = 0
    ds["nx"] = nx
    ds["ny"] = ny
    ds["nz"] = nz
    ds["MZ"] = 1 * lengths[3]
    if squashed:
        ds["MXSUB"] = lengths[1] // nxpe
        ds["MYSUB"] = lengths[2] // nype
        ds["MZSUB"] = lengths[3]
    else:
        ds["MXSUB"] = lengths[1]
        ds["MYSUB"] = lengths[2]
        ds["MZSUB"] = lengths[3]

    MYSUB = lengths[2]

    extra_boundary_points = 0

    if topology == "core":
        ds["ixseps1"] = nx
        ds["ixseps2"] = nx
        ds["jyseps1_1"] = -1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - 1
        ds["ny_inner"] = ny // 2
    elif topology == "sol":
        ds["ixseps1"] = 0
        ds["ixseps2"] = 0
        ds["jyseps1_1"] = -1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - 1
        ds["ny_inner"] = ny // 2
    elif topology == "limiter":
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx
        ds["jyseps1_1"] = -1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - 1
        ds["ny_inner"] = ny // 2
    elif topology == "xpoint":
        if nype < 4 and not squashed:
            raise ValueError(f"Not enough processors for xpoint topology: nype={nype}")
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx // 2
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 2 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = MYSUB - 1
        ds["jyseps1_2"] = ny - MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    elif topology == "single-null":
        if nype < 3 and not squashed:
            raise ValueError(f"Not enough processors for xpoint topology: nype={nype}")
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx
        ds["jyseps1_1"] = MYSUB - 1
        ds["jyseps2_1"] = ny // 2 - 1
        ds["jyseps1_2"] = ny // 2 - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
        ds["ny_inner"] = ny // 2
    elif topology == "connected-double-null":
        if nype < 6 and not squashed:
            raise ValueError(
                "Not enough processors for connected-double-null topology: "
                f"nype={nype}"
            )
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx // 2
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 3 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = ny_inner - MYSUB - 1
        ds["jyseps1_2"] = ny_inner + MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    elif topology == "lower-disconnected-double-null":
        if nype < 6 and not squashed:
            raise ValueError(
                "Not enough processors for lower-disconnected-double-null "
                f"topology: nype={nype}"
            )
        ds["ixseps1"] = nx // 2
        ds["ixseps2"] = nx // 2 + 4
        if ds["ixseps2"] >= nx:
            raise ValueError(
                "Not enough points in the x-direction. ixseps2="
                f'{ds["ixseps2"]} > nx={nx}'
            )
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 3 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = ny_inner - MYSUB - 1
        ds["jyseps1_2"] = ny_inner + MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    elif topology == "upper-disconnected-double-null":
        if nype < 6 and not squashed:
            raise ValueError(
                "Not enough processors for upper-disconnected-double-null "
                f"topology: nype={nype}"
            )
        ds["ixseps2"] = nx // 2
        ds["ixseps1"] = nx // 2 + 4
        if ds["ixseps2"] >= nx:
            raise ValueError(
                "Not enough points in the x-direction. ixseps2="
                f'{ds["ixseps2"]} > nx={nx}'
            )
        ds["jyseps1_1"] = MYSUB - 1
        ny_inner = 3 * MYSUB
        ds["ny_inner"] = ny_inner
        ds["jyseps2_1"] = ny_inner - MYSUB - 1
        ds["jyseps1_2"] = ny_inner + MYSUB - 1
        ds["jyseps2_2"] = ny - MYSUB - 1
    else:
        raise ValueError(f"Unrecognised topology={topology}")

    if metric_3D:
        one = DataArray(np.ones((x_length, y_length, z_length)), dims=["x", "y", "z"])
        zero = DataArray(np.zeros((x_length, y_length, z_length)), dims=["x", "y", "z"])
    else:
        one = DataArray(np.ones((x_length, y_length)), dims=["x", "y"])
        zero = DataArray(np.zeros((x_length, y_length)), dims=["x", "y"])

    ds["zperiod"] = 1
    ds["ZMIN"] = 0.0
    ds["ZMAX"] = 1.0
    ds["g11"] = one
    ds["g22"] = one
    ds["g33"] = one
    ds["g12"] = zero
    ds["g13"] = zero
    ds["g23"] = zero
    ds["g_11"] = one
    ds["g_22"] = one
    ds["g_33"] = one
    ds["g_12"] = zero
    ds["g_13"] = zero
    ds["g_23"] = zero
    ds["G1"] = zero
    ds["G2"] = zero
    ds["G3"] = zero
    ds["J"] = one
    ds["Bxy"] = one
    ds["zShift"] = zero

    ds["dx"] = 0.5 * one
    ds["dy"] = 2.0 * one
    if bout_v5:
        ds["dz"] = 2.0 * one * np.pi / nz
    else:
        ds["dz"] = 2.0 * np.pi / nz

    ds["iteration"] = t_length - 1
    ds["hist_hi"] = t_length - 1
    ds["t_array"] = DataArray(np.arange(t_length, dtype=float) * 10.0, dims="t")
    ds["tt"] = ds["t_array"][-1]

    # xarray adds this encoding when opening a file. Emulate here as it may be used to
    # get the file number
    ds.encoding["source"] = f"BOUT.dmp.{num}.nc"

    _create_bout_ds_cache[call_args] = ds
    return deepcopy(ds)


_create_bout_grid_ds_cache = {}


def create_bout_grid_ds(xsize=2, ysize=4, guards={}, topology="core", ny_inner=0):
    call_args = _get_kwargs()

    try:
        # Has been called with the same signature before, just return the cached result
        return deepcopy(_create_bout_grid_ds_cache[call_args])
    except KeyError:
        pass

    # Set the shape of the data in this dataset
    mxg = guards.get("x", 0)
    myg = guards.get("y", 0)
    xsize += 2 * mxg
    ysize += 2 * myg

    # jyseps* from grid file only ever used to check topology when loading the grid file,
    # so do not need to be consistent with the main dataset
    jyseps2_1 = ysize // 2
    jyseps1_2 = jyseps2_1

    if "double-null" in topology or "xpoint" in topology:
        # Has upper target as well
        ysize += 2 * myg

        # make different from jyseps2_1 so double-null toplogy is recognised
        jyseps1_2 += 1

    shape = (xsize, ysize)

    data = DataArray(np.ones(shape), dims=["x", "y"])

    ds = Dataset(
        {
            "psixy": data,
            "Rxy": data,
            "Zxy": data,
            "hthe": data,
            "y_boundary_guards": myg,
            "jyseps2_1": jyseps2_1,
            "jyseps1_2": jyseps1_2,
            "ny_inner": ny_inner,
            "y_boundary_guards": myg,
        }
    )

    _create_bout_grid_ds_cache[call_args] = ds
    return deepcopy(ds)


# Note, MYPE, PE_XIND and PE_YIND not included, since they are different for each
# processor and so are dropped when loading datasets.
METADATA_VARS = [
    "BOUT_VERSION",
    "NXPE",
    "NYPE",
    "NZPE",
    "MXG",
    "MYG",
    "MZG",
    "nx",
    "ny",
    "nz",
    "MZ",
    "MXSUB",
    "MYSUB",
    "MZSUB",
    "hist_hi",
    "iteration",
    "ixseps1",
    "ixseps2",
    "jyseps1_1",
    "jyseps1_2",
    "jyseps2_1",
    "jyseps2_2",
    "ny_inner",
    "tt",
    "zperiod",
    "ZMIN",
    "ZMAX",
    "use_metric_3d",
]


class TestStripMetadata:
    def test_strip_metadata(self):

        original = create_bout_ds()
        assert original["NXPE"] == 1

        ds, metadata = _separate_metadata(original)

        assert original.drop_vars(
            METADATA_VARS + _BOUT_PER_PROC_VARIABLES + _BOUT_TIME_DEPENDENT_META_VARS,
            errors="ignore",
        ).equals(ds)
        assert metadata["NXPE"] == 1


# TODO also test loading multiple files which have guard cells
class TestOpen:
    def test_single_file(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=1, nype=1, nt=1, write_to_disk=True
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)
        expected = create_bout_ds()
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS
                + _BOUT_PER_PROC_VARIABLES
                + _BOUT_TIME_DEPENDENT_META_VARS,
                errors="ignore",
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(None, nxpe=1, nype=1, nt=1)
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    def test_squashed_file(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=3, nt=1, squashed=True, write_to_disk=True
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)
        expected = create_bout_ds(lengths=(6, 8, 12, 7))
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS
                + _BOUT_PER_PROC_VARIABLES
                + _BOUT_TIME_DEPENDENT_META_VARS,
                errors="ignore",
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(None, nxpe=4, nype=3, nt=1, squashed=True)
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    @pytest.mark.parametrize(
        "keep_xboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    @pytest.mark.parametrize(
        "keep_yboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    def test_squashed_doublenull(
        self,
        tmp_path_factory,
        bout_xyt_example_files,
        keep_xboundaries,
        keep_yboundaries,
    ):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=6,
            nt=1,
            lengths=(6, 2, 4, 7),
            guards={"x": 2, "y": 2},
            squashed=True,
            topology="lower-disconnected-double-null",
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=path,
                keep_xboundaries=keep_xboundaries,
                keep_yboundaries=keep_yboundaries,
            )

        # bout_xyt_example_files when creating a 'squashed' file just makes it with
        # y-size nype*lengths[2]+2*myg, which is 6*4+4=28, so with upper and lower
        # boundaries removed, y-size should be 28-4*2=20.
        assert ds.sizes["t"] == 6
        assert ds.sizes["x"] == 12 if keep_xboundaries else 8
        assert ds.sizes["y"] == 32 if keep_yboundaries else 24
        assert ds.sizes["z"] == 7

    @pytest.mark.parametrize(
        "keep_xboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    @pytest.mark.parametrize(
        "keep_yboundaries", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    def test_squashed_doublenull_file(
        self,
        tmp_path_factory,
        bout_xyt_example_files,
        keep_xboundaries,
        keep_yboundaries,
    ):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=6,
            nt=1,
            lengths=(6, 4, 4, 7),
            guards={"x": 2, "y": 2},
            squashed=True,
            write_to_disk=True,
            topology="upper-disconnected-double-null",
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=path,
                keep_xboundaries=keep_xboundaries,
                keep_yboundaries=keep_yboundaries,
            )

        # bout_xyt_example_files when creating a 'squashed' file just makes it with
        # y-size nype*lengths[2]+2*myg, which is 6*4+4=28, so with upper and lower
        # boundaries removed, y-size should be 28-4*2=20.
        assert ds.sizes["t"] == 6
        assert ds.sizes["x"] == 20 if keep_xboundaries else 16
        assert ds.sizes["y"] == 32 if keep_yboundaries else 24
        assert ds.sizes["z"] == 7

    def test_combine_along_x(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=1,
            nt=1,
            syn_data_type="stepped",
            write_to_disk=True,
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)

        bout_ds = create_bout_ds
        expected = concat(
            [bout_ds(0), bout_ds(1), bout_ds(2), bout_ds(3)],
            dim="x",
            data_vars="minimal",
        )
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS + _BOUT_PER_PROC_VARIABLES, errors="ignore"
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(
            None, nxpe=4, nype=1, nt=1, syn_data_type="stepped"
        )
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    def test_combine_along_y(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=1,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            write_to_disk=True,
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)

        bout_ds = create_bout_ds
        expected = concat(
            [bout_ds(0), bout_ds(1), bout_ds(2)], dim="y", data_vars="minimal"
        )
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(
                METADATA_VARS + _BOUT_PER_PROC_VARIABLES, errors="ignore"
            ),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(
            None, nxpe=1, nype=3, nt=1, syn_data_type="stepped"
        )
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    @pytest.mark.skip
    def test_combine_along_t(self):
        ...

    @pytest.mark.parametrize(
        "bout_v5,metric_3D", [(False, False), (True, False), (True, True)]
    )
    @pytest.mark.parametrize("lengths", [(6, 2, 4, 7), (6, 2, 4, 1)])
    def test_combine_along_xy(
        self, tmp_path_factory, bout_xyt_example_files, bout_v5, metric_3D, lengths
    ):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=3,
            nt=1,
            lengths=lengths,
            syn_data_type="stepped",
            write_to_disk=True,
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )
        with pytest.warns(UserWarning):
            actual = open_boutdataset(datapath=path, keep_xboundaries=False)

        def bout_ds(syn_data_type):
            return create_bout_ds(
                syn_data_type, bout_v5=bout_v5, metric_3D=metric_3D, lengths=lengths
            )

        line1 = concat(
            [bout_ds(0), bout_ds(1), bout_ds(2), bout_ds(3)],
            dim="x",
            data_vars="minimal",
        )
        line2 = concat(
            [bout_ds(4), bout_ds(5), bout_ds(6), bout_ds(7)],
            dim="x",
            data_vars="minimal",
        )
        line3 = concat(
            [bout_ds(8), bout_ds(9), bout_ds(10), bout_ds(11)],
            dim="x",
            data_vars="minimal",
        )
        expected = concat([line1, line2, line3], dim="y", data_vars="minimal")
        expected = expected.set_coords(["t_array", "dx", "dy", "dz"]).rename(
            t_array="t"
        )
        vars_to_drop = METADATA_VARS + _BOUT_PER_PROC_VARIABLES
        xrt.assert_equal(
            actual.drop_vars(["x", "y", "z"]).load(),
            expected.drop_vars(vars_to_drop, errors="ignore"),
        )

        # check creation without writing to disk gives identical result
        fake_ds_list = bout_xyt_example_files(
            None,
            nxpe=4,
            nype=3,
            nt=1,
            lengths=lengths,
            syn_data_type="stepped",
            bout_v5=bout_v5,
            metric_3D=metric_3D,
        )
        with pytest.warns(UserWarning):
            fake = open_boutdataset(datapath=fake_ds_list, keep_xboundaries=False)
        xrt.assert_identical(actual, fake)

    def test_toroidal(self, tmp_path_factory, bout_xyt_example_files):
        # actually write these to disk to test the loading fully
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=3,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            grid="grid",
            write_to_disk=True,
        )
        actual = open_boutdataset(
            datapath=path,
            geometry="toroidal",
            gridfilepath=path.parent.joinpath("grid.nc"),
        )

        # check dataset can be saved
        save_dir = tmp_path_factory.mktemp("data")
        actual.bout.save(save_dir.joinpath("boutdata.nc"))

        # check creation without writing to disk gives identical result
        fake_ds_list, fake_grid_ds = bout_xyt_example_files(
            None,
            nxpe=3,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            grid="grid",
        )
        fake = open_boutdataset(
            datapath=fake_ds_list, geometry="toroidal", gridfilepath=fake_grid_ds
        )
        xrt.assert_identical(actual, fake)

    def test_salpha(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=3,
            nype=3,
            nt=1,
            syn_data_type="stepped",
            grid="grid",
            write_to_disk=True,
        )
        actual = open_boutdataset(
            datapath=path,
            geometry="s-alpha",
            gridfilepath=path.parent.joinpath("grid.nc"),
        )

        # check dataset can be saved
        save_dir = tmp_path_factory.mktemp("data")
        actual.bout.save(save_dir.joinpath("boutdata.nc"))

        # check creation without writing to disk gives identical result
        fake_ds_list, fake_grid_ds = bout_xyt_example_files(
            None, nxpe=3, nype=3, nt=1, syn_data_type="stepped", grid="grid"
        )
        fake = open_boutdataset(
            datapath=fake_ds_list, geometry="s-alpha", gridfilepath=fake_grid_ds
        )
        xrt.assert_identical(actual, fake)

    def test_drop_vars(self, tmp_path_factory, bout_xyt_example_files):
        datapath = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=4,
            nype=1,
            nt=1,
            syn_data_type="stepped",
            write_to_disk=True,
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=datapath, keep_xboundaries=False, drop_variables=["T"]
            )

        assert "T" not in ds.keys()
        assert "n" in ds.keys()

    @pytest.mark.skip
    def test_combine_along_tx(self):
        ...

    def test_restarts(self):
        datapath = Path(__file__).parent.joinpath(
            "data", "restart", "BOUT.restart.*.nc"
        )
        ds = open_boutdataset(datapath, keep_xboundaries=True, keep_yboundaries=True)

        assert "T" in ds


_test_processor_layouts_list = [
    # No parallelization
    (0, 0, 1, 1, {"x": True, "y": True}, {"x": True, "y": True}),
    # 1d parallelization along x:
    # Left
    (0, 0, 3, 1, {"x": True, "y": True}, {"x": False, "y": True}),
    # Middle
    (1, 0, 3, 1, {"x": False, "y": True}, {"x": False, "y": True}),
    # Right
    (2, 0, 3, 1, {"x": False, "y": True}, {"x": True, "y": True}),
    # 1d parallelization along y:
    # Bottom
    (0, 0, 1, 3, {"x": True, "y": True}, {"x": True, "y": False}),
    # Middle
    (0, 1, 1, 3, {"x": True, "y": False}, {"x": True, "y": False}),
    # Top
    (0, 2, 1, 3, {"x": True, "y": False}, {"x": True, "y": True}),
    # 2d parallelization:
    # Bottom left corner
    (0, 0, 3, 4, {"x": True, "y": True}, {"x": False, "y": False}),
    # Bottom right corner
    (2, 0, 3, 4, {"x": False, "y": True}, {"x": True, "y": False}),
    # Top left corner
    (0, 3, 3, 4, {"x": True, "y": False}, {"x": False, "y": True}),
    # Top right corner
    (2, 3, 3, 4, {"x": False, "y": False}, {"x": True, "y": True}),
    # Centre
    (1, 2, 3, 4, {"x": False, "y": False}, {"x": False, "y": False}),
    # Left side
    (0, 2, 3, 4, {"x": True, "y": False}, {"x": False, "y": False}),
    # Right side
    (2, 2, 3, 4, {"x": False, "y": False}, {"x": True, "y": False}),
    # Bottom side
    (1, 0, 3, 4, {"x": False, "y": True}, {"x": False, "y": False}),
    # Top side
    (1, 3, 3, 4, {"x": False, "y": False}, {"x": False, "y": True}),
]

_test_processor_layouts_doublenull_list = [
    # 1d parallelization along y:
    # Bottom
    (0, 0, 1, 4, {"x": True, "y": True}, {"x": True, "y": False}),
    # Lower Middle
    (0, 1, 1, 4, {"x": True, "y": False}, {"x": True, "y": True}),
    # Upper Middle
    (0, 2, 1, 4, {"x": True, "y": True}, {"x": True, "y": False}),
    # Top
    (0, 3, 1, 4, {"x": True, "y": False}, {"x": True, "y": True}),
    # 2d parallelization:
    # Bottom left corner
    (0, 0, 3, 4, {"x": True, "y": True}, {"x": False, "y": False}),
    (1, 0, 3, 4, {"x": False, "y": True}, {"x": False, "y": False}),
    # Bottom right corner
    (2, 0, 3, 4, {"x": False, "y": True}, {"x": True, "y": False}),
    (0, 1, 3, 4, {"x": True, "y": False}, {"x": False, "y": True}),
    (1, 1, 3, 4, {"x": False, "y": False}, {"x": False, "y": True}),
    (2, 1, 3, 4, {"x": False, "y": False}, {"x": True, "y": True}),
    (0, 2, 3, 4, {"x": True, "y": True}, {"x": False, "y": False}),
    (1, 2, 3, 4, {"x": False, "y": True}, {"x": False, "y": False}),
    (2, 2, 3, 4, {"x": False, "y": True}, {"x": True, "y": False}),
    # Top left corner
    (0, 3, 3, 4, {"x": True, "y": False}, {"x": False, "y": True}),
    (1, 3, 3, 4, {"x": False, "y": False}, {"x": False, "y": True}),
    # Top right corner
    (2, 3, 3, 4, {"x": False, "y": False}, {"x": True, "y": True}),
]


class TestTrim:
    @pytest.mark.parametrize("is_restart", [False, True])
    def test_no_trim(self, is_restart):
        ds = create_test_data(0)
        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"
        actual = _trim(
            ds, guards={}, keep_boundaries={}, nxpe=1, nype=1, is_restart=is_restart
        )
        xrt.assert_equal(actual, ds)

    def test_trim_guards(self):
        ds = create_test_data(0)
        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"
        actual = _trim(
            ds, guards={"time": 2}, keep_boundaries={}, nxpe=1, nype=1, is_restart=False
        )
        selection = {"time": slice(2, -2)}
        expected = ds.isel(**selection)
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_list,
    )
    def test_infer_boundaries_2d_parallelization(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        y  9 10 11
        ^  6 7  8
        |  3 4  5
        |  0 1  2
         -----> x
        """

        ds = create_test_data(0)
        ds["jyseps2_1"] = 0
        ds["jyseps1_2"] = 0
        ds["PE_XIND"] = xproc
        ds["PE_YIND"] = yproc
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_doublenull_list,
    )
    def test_infer_boundaries_2d_parallelization_doublenull(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        y  9 10 11
        ^  6 7  8
        |  3 4  5
        |  0 1  2
         -----> x
        """

        ds = create_test_data(0)
        ds["jyseps2_1"] = 3
        ds["jyseps1_2"] = 11
        ds["ny_inner"] = 8
        ds["MYSUB"] = 4
        ds["PE_XIND"] = xproc
        ds["PE_YIND"] = yproc
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_list,
    )
    def test_infer_boundaries_2d_parallelization_by_filenum(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        y  9 10 11
        ^  6 7  8
        |  3 4  5
        |  0 1  2
         -----> x
        """

        filenum = yproc * nxpe + xproc

        ds = create_test_data(0)
        ds["jyseps2_1"] = 0
        ds["jyseps1_2"] = 0
        ds.encoding["source"] = "folder0/BOUT.dmp." + str(filenum) + ".nc"
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize(
        "xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries",
        _test_processor_layouts_doublenull_list,
    )
    def test_infer_boundaries_2d_parallelization_doublenull_by_filenum(
        self, xproc, yproc, nxpe, nype, lower_boundaries, upper_boundaries
    ):
        """
        Numbering scheme for nxpe=3, nype=4

        y  9 10 11
        ^  6 7  8
        |  3 4  5
        |  0 1  2
         -----> x
        """

        filenum = yproc * nxpe + xproc

        ds = create_test_data(0)
        ds["jyseps2_1"] = 3
        ds["jyseps1_2"] = 11
        ds["ny_inner"] = 8
        ds["MYSUB"] = 4
        ds.encoding["source"] = "folder0/BOUT.dmp." + str(filenum) + ".nc"
        actual_lower_boundaries, actual_upper_boundaries = _infer_contains_boundaries(
            ds, nxpe, nype
        )

        assert actual_lower_boundaries == lower_boundaries
        assert actual_upper_boundaries == upper_boundaries

    @pytest.mark.parametrize("is_restart", [False, True])
    def test_keep_xboundaries(self, is_restart):
        ds = create_test_data(0)
        ds = ds.rename({"dim2": "x"})

        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"

        ds["jyseps2_1"] = 8
        ds["jyseps1_2"] = 8

        actual = _trim(
            ds,
            guards={"x": 2},
            keep_boundaries={"x": True},
            nxpe=1,
            nype=1,
            is_restart=is_restart,
        )
        expected = ds  # Should be unchanged
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize("is_restart", [False, True])
    def test_keep_yboundaries(self, is_restart):
        ds = create_test_data(0)
        ds = ds.rename({"dim2": "y"})

        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp.0.nc"

        ds["jyseps2_1"] = 8
        ds["jyseps1_2"] = 8

        actual = _trim(
            ds,
            guards={"y": 2},
            keep_boundaries={"y": True},
            nxpe=1,
            nype=1,
            is_restart=is_restart,
        )
        expected = ds  # Should be unchanged
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize(
        "filenum, lower, upper",
        [(0, True, False), (1, False, True), (2, True, False), (3, False, True)],
    )
    @pytest.mark.parametrize("is_restart", [False, True])
    def test_keep_yboundaries_doublenull_by_filenum(
        self, filenum, lower, upper, is_restart
    ):
        ds = create_test_data(0)
        ds = ds.rename({"dim2": "y"})

        # Manually add filename - encoding normally added by xr.open_dataset
        ds.encoding["source"] = "folder0/BOUT.dmp." + str(filenum) + ".nc"

        ds["jyseps2_1"] = 3
        ds["jyseps1_2"] = 11
        ds["ny_inner"] = 8
        ds["MYSUB"] = 4

        actual = _trim(
            ds,
            guards={"y": 2},
            keep_boundaries={"y": True},
            nxpe=1,
            nype=4,
            is_restart=is_restart,
        )
        expected = ds  # Should be unchanged
        if not lower:
            expected = expected.isel(y=slice(2, None, None))
        if not upper:
            expected = expected.isel(y=slice(None, -2, None))
        xrt.assert_equal(expected, actual)

    @pytest.mark.parametrize("is_restart", [False, True])
    def test_trim_timing_info(self, is_restart):
        ds = create_test_data(0)
        from xbout.load import _BOUT_PER_PROC_VARIABLES

        # remove a couple of entries from _BOUT_PER_PROC_VARIABLES so we test that _trim
        # does not fail if not all of them are present
        _BOUT_PER_PROC_VARIABLES = _BOUT_PER_PROC_VARIABLES[:-2]

        for v in _BOUT_PER_PROC_VARIABLES:
            ds[v] = 42.0
        ds = _trim(
            ds, guards={}, keep_boundaries={}, nxpe=1, nype=1, is_restart=is_restart
        )

        expected = create_test_data(0)
        xrt.assert_equal(ds, expected)


fci_shape = (2, 2, 3, 4)
fci_guards = (2, 2, 0)


@pytest.fixture
def create_example_grid_file_fci(tmp_path_factory):
    """
    Mocks up a FCI-like netCDF file, and return the temporary test
    directory containing them.

    Deletes the temporary directory once that test is done.
    """

    # Create grid dataset
    shape = (fci_shape[1] + 2 * fci_guards[0], *fci_shape[2:])
    arr = np.arange(reduce(operator.mul, shape, 1)).reshape(shape)
    grid = DataArray(data=arr, name="R", dims=["x", "y", "z"]).to_dataset()
    grid["Z"] = DataArray(np.random.random(shape), dims=["x", "y", "z"])
    grid["dy"] = DataArray(np.ones(shape), dims=["x", "y", "z"])
    grid = grid.set_coords(["dy"])

    # Create temporary directory
    save_dir = tmp_path_factory.mktemp("griddata")

    # Save
    filepath = save_dir.joinpath("fci.nc")
    grid.to_netcdf(filepath, engine="netcdf4")

    return filepath


@pytest.fixture
def create_example_files_fci(tmp_path_factory):

    return _bout_xyt_example_files(
        tmp_path_factory,
        lengths=fci_shape,
        nxpe=1,
        nype=1,
        # nt=1,
        guards={a: b for a, b in zip("xyz", fci_guards)},
        syn_data_type="random",
        grid=None,
        squashed=False,
        # topology="core",
        write_to_disk=False,
        bout_v5=True,
        metric_3D=True,
    )



import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import warnings

import animatplot as amp

from .utils import (
    _create_norm,
    _decompose_regions,
    _is_core_only,
    plot_separatrices,
    plot_targets,
)
from matplotlib.animation import PillowWriter


if (
    "pcolor.shading" in matplotlib.rcParams
    and matplotlib.rcParams["pcolor.shading"] == "flat"
):
    # "flat" was the old matplotlib default which discarded the last row and column if
    # X, Y and Z were all equal in size. The new "auto" should be better. Need to set
    # this explicitly because "flat" is the default during a deprecation cycle.
    matplotlib.rcParams["pcolor.shading"] = "auto"


def _add_controls(anim, controls, t_label):
    if controls == "both":
        # Add both time slider and play/pause toggle
        anim.controls(timeline_slider_args={"text": t_label})
    elif controls == "timeline":
        # Add time slider
        anim.timeline_slider(text=t_label)
    elif controls == "toggle":
        # Add play/pause toggle
        anim.toggle()
    elif controls is None or controls == "":
        # Add no controls
        pass
    else:
        raise ValueError(f"Unrecognised value for controls={controls}")


def _parse_coord_option(coord, axis_coords, da):
    if isinstance(axis_coords, dict):
        option_value = axis_coords.get(coord, None)
    else:
        option_value = axis_coords

    if option_value is None:
        c = da[coord]
        if "long_name" in c.attrs:
            label = c.long_name
        else:
            label = coord
        if "units" in c.attrs:
            label = label + f" [{c.units}]"
        return c, label
    elif option_value == "index":
        return np.arange(da.sizes[coord]), f"{coord} index"
    elif isinstance(option_value, str):
        c = da[option_value]
        if "long_name" in c.attrs:
            label = c.long_name
        else:
            label = option_value
        if "units" in c.attrs:
            label = label + f" [{c.units}]"
        return c, label
    else:
        return option_value, None


def _normalise_time_coord(time_values):
    """
    amp.Timeline() does not do a good job of displaying time values that are very small
    (they get rounded to zero). This function scales the time values by a power of 10 to
    get nice values and modifies the units by the scale factor.
    """
    tmax = time_values.max()
    if tmax < 1.0e-2 or tmax > 1.0e6:
        scale_pow = int(np.floor(np.log10(tmax)))
        scale_factor = 10**scale_pow
        time_values = time_values / scale_factor
        suffix = f"e{scale_pow}"
    else:
        suffix = ""

    return time_values, suffix


def animate_poloidal(
    da,
    *,
    ax=None,
    cax=None,
    animate_over=None,
    separatrix=True,
    targets=True,
    add_limiter_hatching=True,
    cmap=None,
    axis_coords=None,
    vmin=None,
    vmax=None,
    logscale=False,
    animate=True,
    save_as=None,
    fps=10,
    controls="both",
    aspect=None,
    extend=None,
    **kwargs,
):
    """
    Make a 2D plot in R-Z coordinates using animatplotlib's Pcolormesh, taking into
    account branch cuts (X-points).

    Parameters
    ----------
    da : xarray.DataArray
        A 2D (x,y) DataArray of data to plot
    ax : Axes, optional
        A matplotlib axes instance to plot to. If None, create a new
        figure and axes, and plot to that
    cax : Axes, optional
        Matplotlib axes instance where the colorbar will be plotted. If None, the default
        position created by matplotlab.figure.Figure.colorbar() will be used.
    animate_over : str, optional
        Dimension over which to animate, defaults to the time dimension
    separatrix : bool, optional
        Add dashed lines showing separatrices
    targets : bool, optional
        Draw solid lines at the target surfaces
    add_limiter_hatching : bool, optional
        Draw hatched areas at the targets
    cmap : matplotlib.colors.Colormap instance, optional
        Colors to use for the plot
    axis_coords : None, str, dict
        Coordinates to use for axis labelling. Only affects time coordinate.
        - None: Use the dimension coordinate for each axis, if it exists.
        - "index": Use the integer index values.
        - dict: keys are dimension names, values set axis_coords for each axis
          separately. Values can be: None, "index", the name of a 1d variable or
          coordinate (which must have the dimension given by 'key'), or a 1d
          numpy array, dask array or DataArray whose length matches the length of
          the dimension given by 'key'.
    vmin : float, optional
        Minimum value for the color scale
    vmax : float, optional
        Maximum value for the color scale
    logscale : bool or float, optional
        If True, default to a logarithmic color scale instead of a linear one.
        If a non-bool type is passed it is treated as a float used to set the linear
        threshold of a symmetric logarithmic scale as
        linthresh=min(abs(vmin),abs(vmax))*logscale, defaults to 1e-5 if True is
        passed.
    animate : bool, optional
        If set to false, do not create the animation, just return the blocks
    save_as : True or str, optional
        If str is passed, save the animation as save_as+'.gif'.
        If True is passed, save the animation with a default name,
        '<variable name>_over_<animate_over>.gif'
    fps : float, optional
        Frame rate for the animation
    controls : string or None, default "both"
        By default, add both the timeline and play/pause toggle to the animation. If
        "timeline" is passed add only the timeline, if "toggle" is passed add only the
        play/pause toggle. If None or an empty string is passed, add neither.
    aspect : str or None, optional
        Argument to set_aspect(), defaults to "equal"
    extend : str or None, optional
        Passed to fig.colorbar()
    **kwargs : optional
        Additional arguments are passed on to the animation method
        animatplot.blocks.Pcolormesh

    Returns
    -------
    animation or blocks
        If animate==True, returns an animatplot.Animation object, otherwise
        returns a list of animatplot.blocks.Pcolormesh instances.
    """

    if animate_over is None:
        animate_over = da.metadata.get("bout_tdim", "t")

    if aspect is None:
        aspect = "equal"

    # TODO generalise this
    x = kwargs.pop("x", "R")
    y = kwargs.pop("y", "Z")

    # Check plot is the right orientation
    spatial_dims = list(da.dims)

    try:
        spatial_dims.remove(animate_over)
    except ValueError:
        raise ValueError(
            "Dimension animate_over={} is not present in the data".format(animate_over)
        )

    if len(da.dims) != 3:
        raise ValueError("da must be 2+1D (t,x,y)")

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    if vmin is None:
        vmin = da.min().values
    if vmax is None:
        vmax = da.max().values

    if extend is None:
        # Replicate default for older matplotlib that does not handle extend=None
        # matplotlib-3.3 definitely does not need this. Not sure about 3.0, 3.1, 3.2.
        extend = "neither"

    # create colorbar
    norm = _create_norm(logscale, kwargs.pop("norm", None), vmin, vmax)
    sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
    sm.set_array([])
    cmap = sm.get_cmap()
    cbar = fig.colorbar(sm, ax=ax, cax=cax, extend=extend)
    if "long_name" in da.attrs:
        cbar_label = da.long_name
    else:
        cbar_label = da.name
    if "units" in da.attrs:
        cbar_label += f" [{da.units}]"
    cbar.ax.set_ylabel(cbar_label)

    ax.set_aspect(aspect)

    da_regions = _decompose_regions(da)

    # Plot all regions on same axis
    blocks = []
    with warnings.catch_warnings():
        # The coordinates we pass are a logically rectangular grid, so should be fine
        # even if this warning is triggered by pcolor or pcolormesh
        warnings.filterwarnings(
            "ignore",
            "The input coordinates to pcolormesh are interpreted as cell centers, but "
            "are not monotonically increasing or decreasing. This may lead to "
            "incorrectly calculated cell edges, in which case, please supply explicit "
            "cell edges to pcolormesh.",
            UserWarning,
        )
        for da_region in da_regions.values():
            # Load values eagerly otherwise for some reason the plotting takes
            # 100's of times longer - for some reason animatplot does not deal
            # well with dask arrays!
            blocks.append(
                amp.blocks.Pcolormesh(
                    da_region.coords[x].values,
                    da_region.coords[y].values,
                    da_region.values,
                    ax=ax,
                    cmap=cmap,
                    norm=norm,
                    **kwargs,
                )
            )

    ax.set_title(da.name)
    ax.set_xlabel(x)
    ax.set_ylabel(y)

    if _is_core_only(da):
        separatrix = False
        targets = False

    if separatrix:
        plot_separatrices(da_regions, ax, x=x, y=y)

    if targets:
        plot_targets(da_regions, ax, x=x, y=y, hatching=add_limiter_hatching)

    if animate:
        t_values, t_label = _parse_coord_option(animate_over, axis_coords, da)
        t_values, t_suffix = _normalise_time_coord(t_values)

        timeline = amp.Timeline(t_values, fps=fps, units=t_suffix)
        anim = amp.Animation(blocks, timeline)

        _add_controls(anim, controls, t_label)

        if save_as is not None:
            if save_as is True:
                save_as = "{}_over_{}".format(da.name, animate_over)
            anim.save(save_as + ".gif", writer=PillowWriter(fps=fps))

        return anim

    return blocks


def animate_pcolormesh(
    data,
    animate_over=None,
    x=None,
    y=None,
    animate=True,
    axis_coords=None,
    vmin=None,
    vmax=None,
    vsymmetric=False,
    logscale=False,
    fps=10,
    save_as=None,
    ax=None,
    cax=None,
    aspect=None,
    extend=None,
    controls="both",
    **kwargs,
):
    """
    Plots a color plot which is animated with time over the specified
    coordinate.

    Currently only supports 2D+1 data, which it plots with animatplotlib's
    wrapping of matplotlib's pcolormesh.

    Parameters
    ----------
    data : xarray.DataArray
    animate_over : str, optional
        Dimension over which to animate, defaults to the time dimension
    x : str, optional
        Dimension to use on the x axis, default is None - then use the first spatial
        dimension of the data
    y : str, optional
        Dimension to use on the y axis, default is None - then use the second spatial
        dimension of the data
    animate : bool, optional
        If set to false, do not create the animation, just return the block
    axis_coords : None, str, dict
        Coordinates to use for axis labelling.
        - None: Use the dimension coordinate for each axis, if it exists.
        - "index": Use the integer index values.
        - dict: keys are dimension names, values set axis_coords for each axis
          separately. Values can be: None, "index", the name of a 1d variable or
          coordinate (which must have the dimension given by 'key'), or a 1d
          numpy array, dask array or DataArray whose length matches the length of
          the dimension given by 'key'.
    vmin : float, optional
        Minimum value to use for colorbar. Default is to use minimum value of
        data across whole timeseries.
    vmax : float, optional
        Maximum value to use for colorbar. Default is to use maximum value of
        data across whole timeseries.
    vsymmetric : bool, optional
        If set to true, make the color-scale symmetric
    logscale : bool or float, optional
        If True, default to a logarithmic color scale instead of a linear one.
        If a non-bool type is passed it is treated as a float used to set the linear
        threshold of a symmetric logarithmic scale as
        linthresh=min(abs(vmin),abs(vmax))*logscale, defaults to 1e-5 if True is
        passed.
    fps : int, optional
        Frames per second of resulting gif
    save_as : True or str, optional
        If str is passed, save the animation as save_as+'.gif'.
        If True is passed, save the animation with a default name,
        '<variable name>_over_<animate_over>.gif'
    ax : Axes, optional
        A matplotlib axes instance to plot to. If None, create a new
        figure and axes, and plot to that
    cax : Axes, optional
        Matplotlib axes instance where the colorbar will be plotted. If None, the default
        position created by matplotlab.figure.Figure.colorbar() will be used.
    aspect : str or None, optional
        Argument to set_aspect(), defaults to "auto"
    extend : str or None, optional
        Passed to fig.colorbar()
    controls : string or None, default "both"
        By default, add both the timeline and play/pause toggle to the animation. If
        "timeline" is passed add only the timeline, if "toggle" is passed add only the
        play/pause toggle. If None or an empty string is passed, add neither.
    kwargs : dict, optional
        Additional keyword arguments are passed on to the animation function
        animatplot.blocks.Pcolormesh

    Returns
    -------
    animation or block
        If animate==True, returns an animatplot.Animation object, otherwise
        returns an animatplot.blocks.Pcolormesh instance.
    """

    if animate_over is None:
        animate_over = data.metadata.get("bout_tdim", "t")

    if aspect is None:
        aspect = "auto"

    variable = data.name

    # Check plot is the right orientation
    spatial_dims = list(data.dims)

    if len(data.dims) != 3:
        raise ValueError("Data passed to animate_imshow must be 3-dimensional")

    try:
        spatial_dims.remove(animate_over)
    except ValueError:
        raise ValueError(
            "Dimension animate_over={} is not present in the data".format(animate_over)
        )

    if x is None and y is None:
        x, y = spatial_dims
    elif x is None:
        try:
            spatial_dims.remove(y)
        except ValueError:
            raise ValueError("Dimension {} is not present in the data".format(y))
        x = spatial_dims[0]
    elif y is None:
        try:
            spatial_dims.remove(x)
        except ValueError:
            raise ValueError("Dimension {} is not present in the data".format(x))
        y = spatial_dims[0]

    if extend is None:
        # Replicate default for older matplotlib that does not handle extend=None
        # matplotlib-3.3 definitely does not need this. Not sure about 3.0, 3.1, 3.2.
        extend = "neither"

    x_values, x_label = _parse_coord_option(x, axis_coords, data)
    y_values, y_label = _parse_coord_option(y, axis_coords, data)

    data = data.transpose(animate_over, y, x, transpose_coords=True)

    # Load values eagerly otherwise for some reason the plotting takes
    # 100's of times longer - for some reason animatplot does not deal
    # well with dask arrays!
    image_data = data.values

    # If not specified, determine max and min values across entire data series
    if vmax is None:
        vmax = np.max(image_data)
    if vmin is None:
        vmin = np.min(image_data)
    if vsymmetric:
        vmax = max(np.abs(vmin), np.abs(vmax))
        vmin = -vmax
    kwargs["norm"] = _create_norm(logscale, kwargs.get("norm", None), vmin, vmax)

    if not ax:
        fig, ax = plt.subplots()

    ax.set_aspect(aspect)

    # Note: animatplot's Pcolormesh gave strange outputs without passing
    # explicitly x- and y-value arrays, although in principle these should not
    # be necessary.
    ny, nx = image_data.shape[1:]
    with warnings.catch_warnings():
        # The coordinates we pass are a logically rectangular grid, so should be fine
        # even if this warning is triggered by pcolor or pcolormesh
        warnings.filterwarnings(
            "ignore",
            "The input coordinates to pcolormesh are interpreted as cell centers, but "
            "are not monotonically increasing or decreasing. This may lead to "
            "incorrectly calculated cell edges, in which case, please supply explicit "
            "cell edges to pcolormesh.",
            UserWarning,
        )
        pcolormesh_block = amp.blocks.Pcolormesh(
            x_values, y_values, image_data, ax=ax, **kwargs
        )

    if animate:
        t_values, t_label = _parse_coord_option(animate_over, axis_coords, data)
        t_values, t_suffix = _normalise_time_coord(t_values)

        timeline = amp.Timeline(t_values, fps=fps, units=t_suffix)
        anim = amp.Animation([pcolormesh_block], timeline)

    cbar = plt.colorbar(pcolormesh_block.quad, ax=ax, cax=cax, extend=extend)
    if "long_name" in data.attrs:
        cbar_label = data.long_name
    else:
        cbar_label = variable
    if "units" in data.attrs:
        cbar_label += f" [{data.units}]"
    cbar.ax.set_ylabel(cbar_label)

    # Add title and axis labels
    ax.set_title(variable)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if animate:
        _add_controls(anim, controls, t_label)

        if save_as is not None:
            if save_as is True:
                save_as = "{}_over_{}".format(variable, animate_over)
            anim.save(save_as + ".gif", writer=PillowWriter(fps=fps))

        return anim

    return pcolormesh_block


def animate_line(
    data,
    animate_over=None,
    animate=True,
    axis_coords=None,
    vmin=None,
    vmax=None,
    fps=10,
    save_as=None,
    sep_pos=None,
    ax=None,
    aspect=None,
    controls="both",
    **kwargs,
):
    """
    Plots a line plot which is animated with time.

    Currently only supports 1D+1 data, which it plots with animatplot's Line animation.

    Parameters
    ----------
    data : xarray.DataArray
    animate_over : str, optional
        Dimension over which to animate, defaults to the time dimension
    animate : bool, optional
        If set to false, do not create the animation, just return the block
    axis_coords : None, str, dict
        Coordinates to use for axis labelling.
        - None: Use the dimension coordinate for each axis, if it exists.
        - "index": Use the integer index values.
        - dict: keys are dimension names, values set axis_coords for each axis
          separately. Values can be: None, "index", the name of a 1d variable or
          coordinate (which must have the dimension given by 'key'), or a 1d
          numpy array, dask array or DataArray whose length matches the length of
          the dimension given by 'key'.
    vmin : float, optional
        Minimum value to use for colorbar. Default is to use minimum value of
        data across whole timeseries.
    vmax : float, optional
        Maximum value to use for colorbar. Default is to use maximum value of
        data across whole timeseries.
    fps : int, optional
        Frames per second of resulting gif
    save_as : True or str, optional
        If str is passed, save the animation as save_as+'.gif'.
        If True is passed, save the animation with a default name,
        '<variable name>_over_<animate_over>.gif'
    sep_pos : int, optional
        Radial position at which to plot the separatrix
    ax : Axes, optional
        A matplotlib axes instance to plot to. If None, create a new
        figure and axes, and plot to that
    aspect : str or None, optional
        Argument to set_aspect(), defaults to "auto"
    controls : string or None, default "both"
        By default, add both the timeline and play/pause toggle to the animation. If
        "timeline" is passed add only the timeline, if "toggle" is passed add only the
        play/pause toggle. If None or an empty string is passed, add neither.
    kwargs : dict, optional
        Additional keyword arguments are passed on to the plotting function
        animatplot.blocks.Line

    Returns
    -------
    animation or block
        If animate==True, returns an animatplot.Animation object, otherwise
        returns an animatplot.blocks.Line instance.
    """

    if animate_over is None:
        animate_over = data.metadata.get("bout_tdim", "t")

    if aspect is None:
        aspect = "auto"

    variable = data.name

    # Check plot is the right orientation
    t_read, x_read = data.dims
    if t_read is animate_over:
        x = x_read
    else:
        data = data.transpose(animate_over, t_read, transpose_coords=True)
        x = t_read

    # Load values eagerly otherwise for some reason the plotting takes
    # 100's of times longer - for some reason animatplot does not deal
    # well with dask arrays!
    image_data = data.values

    # If not specified, determine max and min values across entire data series
    if vmax is None:
        vmax = np.max(image_data)
    if vmin is None:
        vmin = np.min(image_data)

    x_values, x_label = _parse_coord_option(x, axis_coords, data)

    if not ax:
        fig, ax = plt.subplots()

    ax.set_aspect(aspect)

    # set range of plot
    ax.set_ylim([vmin, vmax])

    line_block = amp.blocks.Line(x_values, image_data, ax=ax, **kwargs)

    if animate:
        t_values, t_label = _parse_coord_option(animate_over, axis_coords, data)
        t_values, t_suffix = _normalise_time_coord(t_values)

        timeline = amp.Timeline(t_values, fps=fps, units=t_suffix)
        anim = amp.Animation([line_block], timeline)

    # Add title and axis labels
    ax.set_title(variable)
    ax.set_xlabel(x_label)
    if "long_name" in data.attrs:
        y_label = data.long_name
    else:
        y_label = variable
    if "units" in data.attrs:
        y_label = y_label + f" [{data.units}]"
    ax.set_ylabel(y_label)

    # Plot separatrix
    if sep_pos:
        ax.plot_vline(sep_pos, "--")

    if animate:
        _add_controls(anim, controls, t_label)

        if save_as is not None:
            if save_as is True:
                save_as = "{}_over_{}".format(variable, animate_over)
            anim.save(save_as + ".gif", writer=PillowWriter(fps=fps))

        return anim

    return line_block
